#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - smp - corr
# ==========================================
#
#
#
"""

import traceback
import numpy as np
import copy
import re

from .. import calc
from .sample import Sample


# =======================
# Corr Gain
# =======================
def corr_gain(sample: Sample):
    """Blank Correction"""
    corrGain = True
    gain_corrected = np.zeros([10, len(sample.SequenceName)])
    try:
        for i in range(5):
            f, sf = np.array(sample.TotalParam[126 + i * 2:128 + i * 2])
            sf = f * sf / 100
            gain_corrected[i * 2:2 + i * 2] = calc.corr.gain(*sample.SampleIntercept[i * 2:2 + i * 2], f, sf)
    except Exception as e:
        print(f"Gain correction failed")
        print(traceback.format_exc())
        return
    for i in range(0, 10, 2):
        gain_corrected[i] = [gain_corrected[i][index] if corrGain else j for index, j in enumerate(sample.SampleIntercept[i])]
        gain_corrected[i + 1] = [gain_corrected[i + 1][index] if corrGain else j for index, j in enumerate(sample.SampleIntercept[i + 1])]
    sample.CorrectedValues = copy.deepcopy(gain_corrected)


# =======================
# Corr Blank
# =======================
def corr_blank(sample: Sample):
    """Blank Correction"""
    blank_corrected = np.zeros([10, len(sample.SequenceName)])
    try:
        for i in range(5):
            b, sb = copy.deepcopy(sample.BlankIntercept[i * 2: i * 2 + 2])
            f, sf = np.array(sample.TotalParam[126 + i * 2:128 + i * 2])
            sf = f * sf / 100
            _ = calc.corr.gain(*sample.BlankIntercept[i * 2:2 + i * 2], f, sf)
            for index in range(len(sample.BlankIntercept[i * 2])):
                if sample.TotalParam[111][index]:  # use same parameters to correct blank intercepts
                    b[index] = _[0][index]
                    sb[index] = _[1][index]
            blank_corrected[i * 2:2 + i * 2] = calc.corr.blank(
                *sample.CorrectedValues[i * 2:2 + i * 2], b, sb)
    except Exception as e:
        print(traceback.format_exc())
        raise ValueError('Blank correction error')
    for i in range(0, 10, 2):
        blank_corrected[i] = [blank_corrected[i][index] if sample.TotalParam[102][index] else j for index, j in enumerate(sample.CorrectedValues[i])]
        blank_corrected[i + 1] = [blank_corrected[i + 1][index] if sample.TotalParam[102][index] else j for index, j in enumerate(sample.CorrectedValues[i + 1])]
        blank_corrected[i] = [0 if j < 0 and sample.TotalParam[101][index] else j for index, j in enumerate(blank_corrected[i])]
    sample.BlankCorrected = blank_corrected
    sample.CorrectedValues = copy.deepcopy(sample.BlankCorrected)


# =======================
# Mass Discrimination
# =======================
def corr_massdiscr(sample: Sample):
    """Mass Discrimination Correction"""
    corrMassdiscr = sample.TotalParam[103][0]
    if not corrMassdiscr:
        sample.MassDiscrCorrected = copy.deepcopy(sample.BlankCorrected)
        sample.CorrectedValues = copy.deepcopy(sample.MassDiscrCorrected)
        return
    MASS = sample.TotalParam[71:81]
    mdf_corrected = np.zeros([10, len(sample.SequenceName)])
    try:
        for i in range(5):
            if len(sample.BlankCorrected[i * 2:2 + i * 2]) == 0:
                raise ValueError("sample.BlankCorrected is empty.")
            mdf_corrected[i * 2:2 + i * 2] = calc.corr.discr(
                *sample.BlankCorrected[i * 2:2 + i * 2],
                *sample.TotalParam[69:71], m=MASS[i * 2], m40=MASS[8], isRelative=True,
                method=sample.TotalParam[100][0])
    except Exception as e:
        print(traceback.format_exc())
        raise ValueError(f'Mass discrimination correction error: {e}')
    sample.MassDiscrCorrected = mdf_corrected
    sample.CorrectedValues = copy.deepcopy(sample.MassDiscrCorrected)


# =======================
# Decay correction
# =======================
def corr_decay(sample: Sample):
    """ Ar37 and Ar39 Decay Correction
    Parameters
    ----------
    sample

    Returns
    -------

    """
    decay_corrected = np.zeros([10, len(sample.SequenceName)])
    try:
        irradiation_cycles = [list(filter(None, re.split(r'[DS]', each_step))) for each_step in sample.TotalParam[27]]
        t1 = [re.findall(r"\d+", i) for i in sample.TotalParam[31]]  # t1: experimental times
        t2, t3 = [], []  # t2: irradiation times, t3: irradiation durations
        for each_step in irradiation_cycles:
            t2.append([re.findall(r"\d+", item) for i, item in enumerate(each_step) if i % 2 == 0])
            t3.append([item for i, item in enumerate(each_step) if i % 2 == 1])
        decay_corrected[2:4] = calc.corr.decay(
            *sample.MassDiscrCorrected[2:4], t1, t2, t3, *sample.TotalParam[44:46], isRelative=True)
        decay_corrected[6:8] = calc.corr.decay(
            *sample.MassDiscrCorrected[6:8], t1, t2, t3, *sample.TotalParam[42:44], isRelative=True)
        # Negative number set to zero in decay correction
        decay_corrected[2] = [0 if i < 0 else i for i in decay_corrected[2]]
        decay_corrected[6] = [0 if i < 0 else i for i in decay_corrected[6]]
    except Exception as e:
        print(traceback.format_exc())
        raise ValueError('Decay correction correction error')

    corrDecay37 = sample.TotalParam[104]
    corrDecay39 = sample.TotalParam[105]
    sample.CorrectedValues[2] = [val if corrDecay37[idx] else 0 for idx, val in enumerate(decay_corrected[2])]
    sample.CorrectedValues[3] = [val if corrDecay37[idx] else 0 for idx, val in enumerate(decay_corrected[3])]
    sample.CorrectedValues[6] = [val if corrDecay39[idx] else 0 for idx, val in enumerate(decay_corrected[6])]
    sample.CorrectedValues[7] = [val if corrDecay39[idx] else 0 for idx, val in enumerate(decay_corrected[7])]


# =======================
# Degas Calcium derived 37Ar 36Ar 38Ar 39Ar
# =======================
def calc_degas_ca(sample: Sample):
    """ Degas Pattern for Ca
    Parameters
    ----------
    sample

    Returns
    -------

    """
    corrDecasCa = sample.TotalParam[106]
    # n = len(sample.CorrectedValues[2])
    ar37ca = sample.CorrectedValues[2:4]
    ar39ca = calc.arr.mul_factor(ar37ca, sample.TotalParam[8:10], isRelative=True)
    ar38ca = calc.arr.mul_factor(ar37ca, sample.TotalParam[10:12], isRelative=True)
    ar36ca = calc.arr.mul_factor(ar37ca, sample.TotalParam[12:14], isRelative=True)
    sample.DegasValues[8:10] = copy.deepcopy(ar37ca)
    # sample.DegasValues[22:24] = ar39ca if corrDecasCa else [[0] * n, [0] * n]
    # sample.DegasValues[18:20] = ar38ca if corrDecasCa else [[0] * n, [0] * n]
    # sample.DegasValues[4:6] = ar36ca if corrDecasCa else [[0] * n, [0] * n]
    sample.DegasValues[ 4] = [val if corrDecasCa[idx] else 0 for idx, val in enumerate(ar36ca[0])]
    sample.DegasValues[ 5] = [val if corrDecasCa[idx] else 0 for idx, val in enumerate(ar36ca[1])]
    sample.DegasValues[18] = [val if corrDecasCa[idx] else 0 for idx, val in enumerate(ar38ca[0])]
    sample.DegasValues[19] = [val if corrDecasCa[idx] else 0 for idx, val in enumerate(ar38ca[1])]
    sample.DegasValues[22] = [val if corrDecasCa[idx] else 0 for idx, val in enumerate(ar39ca[0])]
    sample.DegasValues[23] = [val if corrDecasCa[idx] else 0 for idx, val in enumerate(ar39ca[1])]
    sample.PublishValues[1] = copy.deepcopy(ar37ca[0])


# =======================
# Degas Potassium derived 39Ar 38Ar 40Ar
# =======================
def calc_degas_k(sample: Sample):
    """ Degas Pattern for K
    Parameters
    ----------
    sample

    Returns
    -------

    """
    corrDecasK = sample.TotalParam[107]
    set_negative_zero = sample.TotalParam[101]
    # n = len(sample.CorrectedValues[6])
    ar39k = calc.arr.sub(sample.CorrectedValues[6:8], sample.DegasValues[22:24])
    ar39k[0] = [0 if val < 0 and set_negative_zero[idx] else val for idx, val in enumerate(ar39k[0])]
    ar40k = calc.arr.mul_factor(ar39k, sample.TotalParam[14:16], isRelative=True)
    ar38k = calc.arr.mul_factor(ar39k, sample.TotalParam[16:18], isRelative=True)

    sample.PublishValues[3] = copy.deepcopy(ar39k[0])
    sample.DegasValues[20:22] = copy.deepcopy(ar39k)
    # sample.DegasValues[30:32] = ar40k if corrDecasK else [[0] * n, [0] * n]
    # sample.DegasValues[16:18] = ar38k if corrDecasK else [[0] * n, [0] * n]
    sample.DegasValues[16] = [val if corrDecasK[idx] else 0 for idx, val in enumerate(ar38k[0])]
    sample.DegasValues[17] = [val if corrDecasK[idx] else 0 for idx, val in enumerate(ar38k[1])]
    sample.DegasValues[30] = [val if corrDecasK[idx] else 0 for idx, val in enumerate(ar40k[0])]
    sample.DegasValues[31] = [val if corrDecasK[idx] else 0 for idx, val in enumerate(ar40k[1])]


# =======================
# Degas Chlorine derived 36Ar 38Ar
# =======================
def calc_degas_cl(sample: Sample):
    """ Degas Pattern for Cl
    Parameters
    ----------
    sample

    Returns
    -------

    """
    corrDecasCl = sample.TotalParam[108]
    decay_const = sample.TotalParam[46:48]
    cl36_cl38_p = sample.TotalParam[56:58]
    ar38ar36 = sample.TotalParam[4:6]
    stand_time_year = sample.TotalParam[32]
    set_negative_zero = sample.TotalParam[101]
    # ============
    decay_const[1] = [decay_const[0][i] * decay_const[1][i] / 100 for i in
                      range(len(decay_const[0]))]  # convert to absolute error
    cl36_cl38_p[1] = [cl36_cl38_p[0][i] * cl36_cl38_p[1][i] / 100 for i in
                      range(len(cl36_cl38_p[0]))]  # convert to absolute error
    # convert to absolute error
    ar38ar36[1] = [ar38ar36[0][i] * ar38ar36[1][i] / 100 for i in range(len(ar38ar36[0]))]
    # ============
    # 36Ar deduct Ca, that is sum of 36Ara and 36ArCl
    ar36acl = calc.arr.sub(sample.CorrectedValues[0:2], sample.DegasValues[4:6])
    # 38Ar deduct K and Ca, that is sum of 38Ara and 38ArCl
    ar38acl = calc.arr.sub(calc.arr.sub(
        sample.CorrectedValues[4:6], sample.DegasValues[16:18]), sample.DegasValues[18:20])
    for index, item in enumerate(ar36acl[0]):
        if set_negative_zero[index]:
            if item < 0:
                ar36acl[0][index] = 0
            if ar38acl[0][index] < 0:
                ar38acl[0][index] = 0
    try:
        if not corrDecasCl:
            raise ValueError("Do not apply degas chlorine")
        v3 = [cl36_cl38_p[0][i] * (1 - np.exp(-1 * decay_const[0][i] * stand_time_year[i])) for i in
              range(len(stand_time_year))]
        s3 = [pow((cl36_cl38_p[1][i] * (1 - np.exp(-1 * decay_const[0][i] * stand_time_year[i]))) ** 2 +
                  (cl36_cl38_p[0][i] * stand_time_year[i] * (np.exp(-1 * decay_const[0][i] * stand_time_year[i])) *
                   decay_const[1][i]) ** 2, 0.5) for i in range(len(stand_time_year))]
        s3 = [calc.err.div((1, 0), (v3[i], s3[i])) for i in range(len(s3))]
        v3 = [1 / v3[i] for i in range(len(v3))]
        # 36ArCl
        ar36cl = [[], []]
        ar36cl[0] = [(ar36acl[0][i] * ar38ar36[0][i] - ar38acl[0][i]) / (ar38ar36[0][i] - v3[i])
                     for i in range(len(ar36acl[0]))]
        s1 = [(ar36acl[1][i] * ar38ar36[0][i] / (ar38ar36[0][i] - v3[i])) ** 2 for i in range(len(ar36cl[0]))]
        s2 = [(ar38acl[1][i] / (ar38ar36[0][i] - v3[i])) ** 2 for i in range(len(ar36cl[0]))]
        s3 = [(s3[i] * (ar36acl[0][i] * ar38ar36[0][i] - ar38acl[0][i]) / (ar38ar36[0][i] - v3[i]) ** 2) ** 2
              for i in range(len(ar36cl[0]))]
        s4 = [(ar36acl[0][i] / (ar38ar36[0][i] - v3[i]) -
               (ar36acl[0][i] * ar38ar36[0][i] - ar38acl[0][i]) / (ar38ar36[0][i] - v3[i]) ** 2) ** 2 *
              (ar38ar36[1][i]) ** 2 for i in range(len(ar36cl[0]))]
        ar36cl[1] = [pow(s1[i] + s2[i] + s3[i] + s4[i], 0.5) for i in range(len(ar36cl[0]))]

        # 38ArCl
        ar38cl = [[], []]
        ar38cl[1] = [calc.err.mul((ar36cl[0][i], ar36cl[1][i]), (v3[i], s3[i])) for i in range(len(ar36cl[0]))]

        # Negative number set to zero
        ar36cl[0] = [0 if i < 0 and set_negative_zero else i for i in ar36cl[0]]
        # force 36ArCl to zero if 36Ar - 36ArCa - 36Cl < 0
        ar36cl[0] = [0 if ar36acl[0][i] - item < 0 and set_negative_zero else item
                     for i, item in enumerate(ar36cl[0])]

        ar38cl[0] = [ar36cl[0][i] * v3[i] for i in range(len(ar36cl[0]))]

    except Exception as e:
        print('Error in corr Cl: {}, lines: {}'.format(e, e.__traceback__.tb_lineno))
        n = len(ar36acl[0])
        ar36cl = np.zeros([2, n])
        ar38cl = np.zeros([2, n])

    # sample.DegasValues[6:8] = ar36cl
    # sample.DegasValues[10:12] = ar38cl
    sample.PublishValues[2] = copy.deepcopy(ar38cl[0])
    sample.DegasValues[ 6] = [val if corrDecasCl[idx] else 0 for idx, val in enumerate(ar36cl[0])]
    sample.DegasValues[ 7] = [val if corrDecasCl[idx] else 0 for idx, val in enumerate(ar36cl[1])]
    sample.DegasValues[10] = [val if corrDecasCl[idx] else 0 for idx, val in enumerate(ar38cl[0])]
    sample.DegasValues[11] = [val if corrDecasCl[idx] else 0 for idx, val in enumerate(ar38cl[1])]


# =======================
# Degas atmospheric 36Ar 38Ar 40Ar
# =======================
def calc_degas_atm(sample: Sample):
    """ Degas for Atmospheric Gas
    Parameters
    ----------
    sample

    Returns
    -------

    """
    corrDecasAtm = sample.TotalParam[109]
    set_negative_zero = sample.TotalParam[101]
    # n = len(sample.CorrectedValues[0])
    # 36Ar deduct Ca, that is sum of 36Ara and 36ArCl
    ar36acl = calc.arr.sub(sample.CorrectedValues[0:2], sample.DegasValues[4:6])
    ar36acl[0] = [0 if val < 0 and set_negative_zero[idx] else val for idx, val in enumerate(ar36acl[0])]
    # 38Ar deduct K and Ca, that is sum of 38Ara and 38ArCl
    # ar38acl = calc.arr.sub()(
    #     calc.arr.sub()(sample.CorrectedValues[2:4], sample.DegasValues[16:18]), sample.DegasValues[18:20])
    # 36ArAir
    ar36a = calc.arr.sub(ar36acl, sample.DegasValues[6:8])
    # If ar36acl - ar36cl < 0, let ar36a = ar36 - ar36ca
    ar36a[0] = [ar36acl[index] if item < 0 and set_negative_zero[index] else item for index, item in enumerate(ar36a[0])]
    if sample.Info.sample.type == "Air":
        ar38a = copy.deepcopy(sample.CorrectedValues[4:6])
        ar40a = copy.deepcopy(sample.CorrectedValues[8:10])
    else:
        # 38ArAir
        ar38a = calc.arr.mul_factor(ar36a, sample.TotalParam[4:6], isRelative=True)
        # 40ArAir
        ar40a = calc.arr.mul_factor(ar36a, sample.TotalParam[0:2], isRelative=True)

    sample.PublishValues[0] = copy.deepcopy(ar36a[0])
    # sample.DegasValues[12:14] = ar38a if corrDecasAtm else [[0] * n, [0] * n]
    # sample.DegasValues[26:28] = ar40a if corrDecasAtm else [[0] * n, [0] * n]
    sample.DegasValues[ 0] = [val if corrDecasAtm[idx] else 0 for idx, val in enumerate(ar36a[0])]
    sample.DegasValues[ 1] = [val if corrDecasAtm[idx] else 0 for idx, val in enumerate(ar36a[1])]
    sample.DegasValues[12] = [val if corrDecasAtm[idx] else 0 for idx, val in enumerate(ar38a[0])]
    sample.DegasValues[13] = [val if corrDecasAtm[idx] else 0 for idx, val in enumerate(ar38a[1])]
    sample.DegasValues[26] = [val if corrDecasAtm[idx] else 0 for idx, val in enumerate(ar40a[0])]
    sample.DegasValues[27] = [val if corrDecasAtm[idx] else 0 for idx, val in enumerate(ar40a[1])]


# =======================
# Degas radiogenic 40Ar
# =======================
def calc_degas_r(sample: Sample):
    """ Degas for Radiogenic Ar40
    Parameters
    ----------
    sample

    Returns
    -------

    """
    ar40ar = calc.arr.sub(sample.CorrectedValues[8:10], sample.DegasValues[30:32])
    ar40r = calc.arr.sub(ar40ar, sample.DegasValues[26:28])
    ar40r[0] = [0 if item < 0 and sample.TotalParam[101][index] else item for index, item in enumerate(ar40r[0])]
    sample.DegasValues[24:26] = copy.deepcopy(ar40r)
    sample.PublishValues[4] = copy.deepcopy(ar40r[0])


# =======================
# Calc ratio
# =======================
def calc_nor_inv_isochrons(sample: Sample):
    n = len(sample.SequenceName)
    try:
        isochron_1 = calc.isochron.get_data(*sample.DegasValues[20:22], *calc.arr.sub(
            sample.CorrectedValues[8:10], sample.DegasValues[30:32]), *sample.DegasValues[0:2])
        isochron_2 = calc.isochron.get_data(*sample.DegasValues[20:22], *sample.DegasValues[0:2], *calc.arr.sub(
            sample.CorrectedValues[8:10], sample.DegasValues[30:32]))
    except:
        return np.zeros([5, n]), np.zeros([5, n])
    else:
        return isochron_1, isochron_2


def calc_cl_isochrons(sample: Sample):
    n = len(sample.SequenceName)
    try:
        isochron_3 = calc.isochron.get_data(
            *sample.DegasValues[20:22], *sample.DegasValues[24:26], *sample.DegasValues[10:12])
        isochron_4 = calc.isochron.get_data(
            *sample.DegasValues[20:22], *sample.DegasValues[10:12], *sample.DegasValues[24:26])
        isochron_5 = calc.isochron.get_data(
            *sample.DegasValues[10:12], *sample.DegasValues[24:26], *sample.DegasValues[20:22])
    except:
        return np.zeros([5, n]), np.zeros([5, n]), np.zeros([5, n])
    else:
        return isochron_3, isochron_4, isochron_5


def calc_3D_isochrons(sample: Sample):
    n = len(sample.SequenceName)
    try:
        # === Ar values ===
        # 3D ratio, 36Ar(a+cl)/40Ar(a+r), 38Ar(a+cl)/40Ar(a+r), 39Ar(k)/40Ar(a+r),
        ar40ar = calc.arr.sub(sample.CorrectedValues[8:10], sample.DegasValues[30:32])
        # 36Ar deduct Ca, that is sum of 36Ara and 36ArCl (and also 36Arc)
        ar36acl = calc.arr.sub(sample.CorrectedValues[0:2], sample.DegasValues[4:6])
        # 38Ar deduct K and Ca, that is sum of 38Ara and 38ArCl (and also 38Arc)
        ar38acl = calc.arr.sub(calc.arr.sub(sample.CorrectedValues[4:6], sample.DegasValues[16:18]),
                               sample.DegasValues[18:20])
        # 38ArCl
        ar38cl = sample.DegasValues[10:12]
        # 39ArK
        ar39k = sample.DegasValues[20:22]
        # isochron_6 = calc.isochron.get_3d_data(*ar36acl, *ar38acl, *ar40ar, *ar39k)
        isochron_6 = calc.isochron.get_3d_data(*ar36acl, *ar38acl, *ar39k, *ar40ar)  # Points on the plot will be more disperse than the above
    except:
        return np.zeros([9, n])
    else:
        return isochron_6


def calc_ratio(sample: Sample, monte_carlo: bool = False):
    """ Calculate isochron ratio data, 40Arr/39ArK, Ar40r percentage,
        Ar39K released percentage, Ca/K
    Parameters
    ----------
    sample : Sample instance
    monte_carlo : whether conduct monte carlo simulation for calculating 40Arr/39ArK

    Returns
    -------
    None
    """
    ar40r_percent = [item / sample.CorrectedValues[8][index] * 100 if sample.CorrectedValues[8][index] != 0 else 0
                     for index, item in enumerate(sample.DegasValues[24])]
    sum_ar39k = sum(sample.DegasValues[20])
    sum_ar36a = sum(sample.DegasValues[ 0])
    CaK = calc.arr.mul_factor(calc.arr.mul_factor(
        sample.DegasValues[8:10], calc.arr.rec_factor(sample.DegasValues[20:22], isRelative=False)),
        calc.arr.rec_factor(sample.TotalParam[20:22], isRelative=True))

    # assignation
    if sample.Info.sample.type == "Air":
        ar40aar36a = calc.arr.mul_factor(
            sample.DegasValues[26:28], calc.arr.rec_factor(sample.DegasValues[0:2], isRelative=False),
            isRelative=False)
        ar36a_percent = [item / sum_ar36a * 100 if sum_ar36a != 0 else 0 for item in sample.DegasValues[0]]
        sample.ApparentAgeValues[0:2] = ar40aar36a
        sample.ApparentAgeValues[7] = ar36a_percent
    else:
        ar40rar39k = calc.arr.mul_factor(
            sample.DegasValues[24:26], calc.arr.rec_factor(sample.DegasValues[20:22], isRelative=False),
            isRelative=False)
        ar39k_percent = [item / sum_ar39k * 100 if sum_ar39k != 0 else 0 for item in sample.DegasValues[20]]
        sample.ApparentAgeValues[0:2] = ar40rar39k
        sample.ApparentAgeValues[7] = ar39k_percent

    sample.ApparentAgeValues[6] = ar40r_percent
    sample.PublishValues[7:11] = copy.deepcopy([*sample.ApparentAgeValues[6:8], *CaK])

    sample.IsochronValues[0:5], sample.IsochronValues[6:11] = calc_nor_inv_isochrons(sample)
    sample.IsochronValues[12:17], sample.IsochronValues[18:23], sample.IsochronValues[24:29] = \
        calc_cl_isochrons(sample)

    # === Cl-Atm-Correlation Plot ===
    sample.IsochronValues[30:39] = calc_3D_isochrons(sample)

    # Turner 1988 3D cake mix plots
    # ar40 = sample.CorrectedValues[8:10]  # ar40 = atm + r + k
    # ar36a = sample.DegasValues[0:2]  # ar36a
    # isochron_6 = calc.isochron.get_3d_data(*ar39k, *ar38cl, *ar40, *ar36a)
    # sample.IsochronValues[30:39] = isochron_6

    # Note that the difference between Turner 3D plots and our 3D plots.

    if monte_carlo:
        res = monte_carlo_f(sample=sample)
        # where to display simulation results
        res = calc.arr.transpose(list(res))  # res is a generator for [age, sage, ...]
        age, sage = res[:2]
        sample.ApparentAgeValues[1] = sage

        # isochron data
        isochron_1 = res[2:7]
        isochron_2 = res[7:12]
        sample.IsochronValues[0:5] = isochron_1
        sample.IsochronValues[6:11] = isochron_2


def monte_carlo_f(sample: Sample):
    """
    Parameters
    ----------
    sample

    Returns
    -------

    """

    sequence_num = sample.sequence().size

    ar40m = np.transpose(sample.SampleIntercept[8:10])
    ar39m = np.transpose(sample.SampleIntercept[6:8])
    ar38m = np.transpose(sample.SampleIntercept[4:6])
    ar37m = np.transpose(sample.SampleIntercept[2:4])
    ar36m = np.transpose(sample.SampleIntercept[0:2])
    ar40b = np.transpose(sample.BlankIntercept[8:10])
    ar39b = np.transpose(sample.BlankIntercept[6:8])
    ar38b = np.transpose(sample.BlankIntercept[4:6])
    ar37b = np.transpose(sample.BlankIntercept[2:4])
    ar36b = np.transpose(sample.BlankIntercept[0:2])

    M36 = np.transpose(sample.TotalParam[71:73])
    M37 = np.transpose(sample.TotalParam[73:75])
    M38 = np.transpose(sample.TotalParam[75:77])
    M39 = np.transpose(sample.TotalParam[77:79])
    M40 = np.transpose(sample.TotalParam[79:81])

    G36 = np.transpose(sample.TotalParam[126:128])  # gain correction factors
    G37 = np.transpose(sample.TotalParam[128:130])
    G38 = np.transpose(sample.TotalParam[130:132])
    G39 = np.transpose(sample.TotalParam[132:134])
    G40 = np.transpose(sample.TotalParam[134:136])

    MDF = np.transpose(sample.TotalParam[69:71])

    L39ar = np.transpose(sample.TotalParam[42:44])
    L37ar = np.transpose(sample.TotalParam[44:46])
    L36cl = np.transpose(sample.TotalParam[46:48])

    R39v37ca = np.transpose(sample.TotalParam[8:10])
    R38v37ca = np.transpose(sample.TotalParam[10:12])
    R36v37ca = np.transpose(sample.TotalParam[12:14])
    R40v39k = np.transpose(sample.TotalParam[14:16])
    R38v39k = np.transpose(sample.TotalParam[16:18])

    R40v36a = np.transpose(sample.TotalParam[0:2])
    R38v36a = np.transpose(sample.TotalParam[4:6])
    R36v38clp = np.transpose(sample.TotalParam[56:58])

    stand_time_year = np.transpose(sample.TotalParam[32])

    irradiation_cycles = [list(filter(None, re.split(r'[DS]', each_step))) for each_step in sample.TotalParam[27]]
    t1 = [re.findall(r"\d+", i) for i in sample.TotalParam[31]]  # t1: experimental times
    t2, t3 = [], []  # t2: irradiation times, t3: irradiation durations
    for each_step in irradiation_cycles:
        t2.append([re.findall(r"\d+", item) for i, item in enumerate(each_step) if i % 2 == 0])
        t3.append([item for i, item in enumerate(each_step) if i % 2 == 1])

    # for i in range(sequence_num):
    #     P37Decay = calc.corr.get_decay_factor(t1[i], t2[i], t3[i], L37ar[i][0], L37ar[i][0] * L37ar[i][1] / 100)
    #     print(P37Decay)
    #
    # for i in range(sequence_num):
    #     P39Decay = calc.corr.get_decay_factor(t1[i], t2[i], t3[i], L39ar[i][0], L39ar[i][0] * L39ar[i][1] / 100)
    #     print(P39Decay)

    for i in range(sequence_num):

        print(f"Monte Carlo Simulation For sequence {i + 1}")

        res = calc.corr.Monte_Carlo_F(
            ar40m=ar40m[i], ar39m=ar39m[i], ar38m=ar38m[i], ar37m=ar37m[i], ar36m=ar36m[i],
            ar40b=ar40b[i], ar39b=ar39b[i], ar38b=ar38b[i], ar37b=ar37b[i], ar36b=ar36b[i],
            M40=M40[i], M39=M39[i], M38=M38[i], M37=M37[i], M36=M36[i],
            G40=G40[i], G39=G39[i], G38=G38[i], G37=G37[i], G36=G36[i],
            t1=t1[i], t2=t2[i], t3=t3[i],
            R40v36a=R40v36a[i], R38v36a=R38v36a[i],
            R39v37ca=R39v37ca[i], R36v37ca=R36v37ca[i], R38v37ca=R38v37ca[i],
            R40v39k=R40v39k[i], R38v39k=R38v39k[i],
            R36v38clp=R36v38clp[i],
            L37ar=L37ar[i], L39ar=L39ar[i], L36cl=L36cl[i],
            MDFunc=None,
            MDF=MDF[i], stand_time_year=stand_time_year[i],
            blank_gain_corr=sample.TotalParam[111][i]
        )

        yield res



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang 
# ararpy - raw
# ==========================================
#
#
# 
"""
import os
import pandas as pd
from typing import List, Union, Optional
from types import MethodType
from ..calc import arr, raw_funcs
from ..files import raw_file
from ..files.basic import read as read_params
from . import (sample as samples)

RawData = samples.RawData
Sequence = samples.Sequence


def to_raw(file_path: Union[str, List[str]], input_filter_path: Union[str, List[str]], **kwargs):
    """ Read raw data from files, can create raw data instance based on the given files
    Raw data will have the structure like:
        [
            [ [ sequence 0 header ], [ measurement cycle 0 data ], [ measurement cycle 1 data ], ...  ],
            [ [ sequence 1 header ], [ measurement cycle 0 data ], [ measurement cycle 1 data ], ...  ],
            ...
            [ [ sequence n - 1 header ], [ measurement cycle 0 data ], [ measurement cycle 1 data ], ...  ],
        ]
    Parameters
    ----------
    file_path
    input_filter_path
    kwargs

    Returns
    -------

    """
    if isinstance(file_path, list) and isinstance(input_filter_path, list):
        raw = concatenate([to_raw(file, input_filter_path[index]) for index, file in enumerate(file_path)])
    elif isinstance(file_path, str) and isinstance(input_filter_path, str):
        input_filter = read_params(input_filter_path)
        file_name = str(os.path.split(file_path)[-1]).split('.')[0]
        res = raw_file.open_file(file_path, input_filter)
        data = res.get('data', None)
        sequences = res.get('sequences', None)
        sequence_num = len(data) if data is not None else len(sequences)
        # default fitting method, all exponential
        # <option value=0>Linear</option>
        # <option value=1>Quadratic</option>
        # <option value=2>Exponential</option>
        # <option value=3>Power</option>
        # <option value=4>Average</option>
        fitting_method = [2, 0, 2, 2, 2]
        raw = RawData(name=file_name, data=data, isotopic_num=10, sequence_num=sequence_num, source=[file_path],
                      sequence=sequences, unit=str(input_filter[30]), fitting_method=[*fitting_method])
    else:
        raise ValueError("File path and input filter should be both string or list with a same length.")
    return raw


def concatenate(raws: List[RawData]):
    """
    Parameters
    ----------
    raws

    Returns
    -------

    """
    step_names = []

    def resort_sequence(seq: Sequence, index):
        count = 0
        while seq.name in step_names:
            # rename
            if count == 0:
                seq.name = f"{seq.name}-{index + 1}"
            else:
                seq.name = f"{seq.name} ({count})"
            count = count + 1
        step_names.append(seq.name)
        seq.index = index
        seq.is_removed = hasattr(seq, "is_removed") and seq.is_removed is True
        return seq

    source = [_source for _raw in raws for _source in _raw.source]
    unit = set([_raw.unit for _raw in raws])
    unit = list(unit)[0] if len(unit) == 1 else 'Unknown Unit'
    sequence = [resort_sequence(seq, index) for index, seq in enumerate([i for _raw in raws for i in _raw.sequence])]
    sequence_num = len(sequence)
    exp_names = list(set([_raw.name for _raw in raws]))
    return RawData(name='&'.join(exp_names), source=source, isotopic_num=10,
                   sequence_num=sequence_num, sequence=sequence, unit=unit)


def get_sequence(raw: RawData, index: Optional[Union[list, int, str, bool]] = None,
                 flag: Optional[str] = None, unique: Optional[bool] = True):
    """
    Parameters
    ----------
    raw
    index :
        value
    flag :
        name of attribution to be matched of a sequence
    unique : bool, if True, will return the first matched sequence,
        False, return a list of all matched sequences

        a = raw.get_sequence(True, 'is_unknown', unique=False)  # get unknown sequence
        print([_a.name for _a in a])

    Returns
    -------

    """
    if index is None:
        return raw.sequence
    if isinstance(index, list):
        return [get_sequence(raw, i, flag) for i in index]
    # judge boolean before int
    if isinstance(index, (str, bool)) and flag is not None:
        return arr.filter(raw.sequence, lambda seq: getattr(seq, flag)() == index if type(
            getattr(seq, flag)) is MethodType else getattr(seq, flag) == index, unique=unique, get=None)
    if isinstance(index, int):
        return raw.sequence[index]


def do_regression(raw: RawData, sequence_index: Optional[List] = None, isotopic_index: Optional[List] = None):
    """
    Parameters
    ----------
    raw
    sequence_index
    isotopic_index

    Returns
    -------

    """

    for sequence in raw.get_sequence(index=None, flag=None):
        if hasattr(sequence_index, '__getitem__') and sequence.index not in sequence_index:
            continue
        isotope: pd.DataFrame = sequence.get_data_df()
        selected: pd.DataFrame = isotope[sequence.get_flag_df()[list(range(1, 11))]]
        # unselected: pd.DataFrame = isotope[~sequence.get_flag_df()[list(range(1, 11))]]
        selected: list = [selected[[isotopic_index * 2 + 1, 2 * (isotopic_index + 1)]].dropna().values.tolist()
                          for isotopic_index in list(range(5))]
        # unselected: list = [unselected[[isotopic_index*2 + 1, 2 * (isotopic_index + 1)]].dropna().values.tolist()
        #                     for isotopic_index in list(range(5))]

        for index, isotope in enumerate(selected):
            if hasattr(isotopic_index, '__getitem__') and index not in isotopic_index:
                continue
            # print(f"regression for {sequence.name = }, isotope {index = }")
            res = raw_funcs.get_raw_data_regression_results(isotope)
            try:
                sequence.results[index] = res[1]
                sequence.coefficients[index] = res[2]
            except IndexError:
                sequence.results.insert(index, res[1])
                sequence.coefficients.insert(index, res[2])
            except TypeError:
                sequence.results = [res[1]]
                sequence.coefficients = [res[2]]



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

#  Copyright (C) 2023 Yang. - All Rights Reserved

import os
import ararpy as ap

example_dir = os.path.join(os.path.dirname(ap.__file__), r'examples')
# example_dir = os.path.join(os.getcwd(), r'examples')

ap.test.test()

# sample = ap.from_empty()  # create new sample instance
# print(sample.show_data())
# # Sample Name:
# #
# # Doi:
# #     9a43b5c1a99747ee8608676ac31814da
# # Corrected Values:
# #     Empty DataFrame
# # Columns: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
# # Index: []
# # Parameters:
# #     Empty DataFrame
# # Columns: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
# #           20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39,
# #           40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59,
# #           60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79,
# #           80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, ...]
# # Index: []
# #
# # [0 rows x 123 columns]
# # Isochron Values:
# #     Empty DataFrame
# # Columns: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
# #           20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39,
# #           40, 41, 42, 43, 44, 45, 46]
# # Index: []
# # Apparent Ages:
# #     Empty DataFrame
# # Columns: [0, 1, 2, 3, 4, 5, 6, 7]
# # Index: []
# # Publish Table:
# #     Empty DataFrame
# # Columns: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
# # Index: []
#
# file_path = os.path.join(example_dir, r'22WHA0433.arr')
# sample = ap.from_arr(file_path)
# # normal isochron age
# print(f"{sample.results().isochron.inverse.set1.age = }")
# # sample.results().isochron.inverse.set1.age = 163.10336210925516
# # check current data point selection
# print(f"{sample.sequence().mark.value}")
# # [nan, nan, nan, nan, nan, nan, nan, nan, nan, nan, nan, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
# #  1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
# print(f"{sample.sequence().mark.set1.index}")
# # [11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26]
#
# # change data point selection
# sample.set_selection(10, 1)
# # check new data point selection
# print(f"{sample.sequence().mark.value}")
# # ['', '', '', '', '', '', '', '', '', '', 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
# print(f"{sample.sequence().mark.set1.index}")
# # [10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26]
# # recalculate
# sample.recalculate(re_plot=True)
# # check new results
# print(f"{sample.results().isochron.inverse.set1.age = }")
# # sample.results().isochron.inverse.set1.age = 164.57644271385772


""" Open ArArCalc .age files """
# file_path = os.path.join(example_dir, r'22WHA0433.age')
# # file_path = os.path.join(example_dir, r'20WHA0103.age')
# sample: ap.Sample = ap.from_age(file_path=file_path)
# print(sample.name())
# print(ap.smp.basic.get_dict_from_obj(sample.Info))
# print(ap.smp.basic.get_dict_from_obj(sample.results()))
#
# ap.smp.corr.monte_carlo_f(sample)

# """ Open ArArCalc .full.xls files """
# file_path = os.path.join(example_dir, r'22WHA0433.full.xls')
# sample = ap.from_full(file_path=file_path)
# print(sample.name())
# print(ap.smp.basic.get_dict_from_obj(sample.Info))
#
# """ Open .arr files """
# file_path = os.path.join(example_dir, r'22WHA0433.arr')
# sample = ap.from_arr(file_path=file_path)
# print(sample.name())
# print(ap.smp.basic.get_dict_from_obj(sample.Info))
#
# # might not correct, because in this way the split numbers and the splits are not calculated correclty.
# # With WebArAr, due to the auto calculation by Echart this issue doesn't exist.
# sample.to_pdf(file_path="myFigure.pdf", figure="figure_3")

""" Open raw files """
# file_path = os.path.join(example_dir, r'22WHA0078.xls')
# input_filter_path = os.path.join(example_dir, r'Qtegra-exported-xls.input-filter')
# file_path = os.path.join(example_dir, r'24WHN0001-51-592.XLS')
# input_filter_path = os.path.join(example_dir, r'NGX-XLS.input-filter')

# sample = ap.from_raw(file_path, input_filter_path)
# print(sample.name())
# print(sample.parameters().to_df())
#
# # set info
# info = {
#     'sample': {'name': '22WHA0078', 'material': 'muscovite', 'location': 'China'},
#     'researcher': {'name': 'Yang'},
#     'laboratory': {'name': 'VU Ar lab', 'info': '', 'analyst': 'Yang'}
# }
# sample.set_info(info=info)
# # set calc parameters
# calc_params_file_path = os.path.join(example_dir, r'ArAr.calc')
# irra_params_file_path = os.path.join(example_dir, r'WH01.irra')
# # param = ap.files.basic.read(calc_params_file_path)
#
# sample.set_params(calc_params_file_path)
# sample.set_params(irra_params_file_path)
#
# print(sample.name())
# print(sample.parameters().to_df())


# """ progress """
# print(sample.sequence().mark.set1.index)
# print(sample.results().isochron.normal.set1.age)
# print(sample.results().age_plateau.set1.age)
# sample.set_selection(8, 1)
# print(sample.sequence().mark.set1.index)
# print(sample.results().isochron.normal.set1.age)
# print(sample.results().age_plateau.set1.age)
# sample.plot_normal()
# sample.plot_inverse()
# sample.plot_age_plateau()
# print(sample.sequence().mark.set1.index)
# print(sample.results().isochron.normal.set1.age)
# print(sample.results().age_plateau.set1.age)
# print(sample.help)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# webarar - raw_file
# ==========================================
#
#
#
"""

from typing import List, Union
import traceback
import os
import re
import pickle
import chardet
from xlrd import open_workbook
from datetime import datetime
from parse import parse as string_parser
import dateutil.parser as datetime_parser
from ..calc.arr import get_item
from ..calc.basic import utc_dt

""" Open raw data file """

DEFAULT_SAMPLE_INFO = {}


def open_file(file_path: str, input_filter: List[Union[str, int, bool]]):
    """
    Parameters
    ----------
    file_path:
    input_filter

    Returns
    -------
    step_list -> [[[header of step one], [cycle one in the step], [cycle two in the step]],[[],[]]]
        example:
            [
                [
                    [1, '6/8/2019  8:20:51 PM', 'BLK', 'B'],  # step header
                    # [sequence index, date time, label, value]
                    ['1', 12.30, 87.73, 12.30, 1.30, 12.30, 0.40, 12.30, 0.40, 12.30, 0.44],  # step/sequence 1
                    # [cycle index, time, Ar40 signal, time, Ar39 signal, ..., time, Ar36 signal]
                    ['2', 24.66, 87.70, 24.66, 1.14, 24.66, 0.36, 24.66, 0.35, 24.66, 0.43],  # step/sequence 2
                    ...
                    ['10', 123.06, 22262.68, 123.06, 6.54, 123.06, 8.29, 123.06, 0.28, 123.06, 29.22],
                ],
                [
                    ...
                ]
            ]

    """
    extension = str(os.path.split(file_path)[-1]).split('.')[-1]
    try:
        handler = {'txt': open_raw_txt, 'excel': open_raw_xls,
                   'Qtegra Exported XLS': open_qtegra_exported_xls, 'Seq': open_raw_seq}[
            ['txt', 'excel', 'Qtegra Exported XLS', 'Seq'][int(input_filter[1])]]
    except KeyError:
        print(traceback.format_exc())
        raise FileNotFoundError("Wrong File.")
    return handler(file_path, input_filter)


def open_qtegra_exported_xls(filepath, input_filter=None):
    if input_filter is None:
        input_filter = []
    try:
        wb = open_workbook(filepath)
        sheets = wb.sheet_names()
        sheet = wb.sheet_by_name(sheets[0])
        value, step_header, step_list = [], [], []
        for row in range(sheet.nrows):
            row_set = []
            for col in range(sheet.ncols):
                if sheet.cell(row, col).value == '':
                    pass
                else:
                    row_set.append(sheet.cell(row, col).value)
            if row_set != [] and len(row_set) > 1:
                value.append(row_set)
        for each_row in value:
            # if the first item of each row is float (1.0, 2.0, ...) this row is the header of a step.
            if isinstance(each_row[0], float):
                each_row[0] = int(each_row[0])
                if "M" in each_row[1].upper():
                    each_row[1] = datetime.strptime(each_row[1], '%m/%d/%Y  %I:%M:%S %p').isoformat(timespec='seconds')
                else:
                    each_row[1] = datetime.strptime(each_row[1], '%m/%d/%Y  %H:%M:%S').isoformat(timespec='seconds')
                step_header.append(each_row)
        for step_index, each_step_header in enumerate(step_header):
            row_start_number = value.index(each_step_header)
            try:
                row_stop_number = value.index(step_header[step_index + 1])
            except IndexError:
                row_stop_number = len(value) + 1
            step_values = [
                each_step_header[0:4],
                *list(map(
                    # lambda x: [x[0], x[1], x[2], x[1], x[3], x[1], x[4], x[1], x[5], x[1], x[6]],
                    # x[1] = time, x[2] = H2:40, x[3] = H1: 39, x[4] = AX: 38, x[5] = L1: 37, x[6] = L2: 36
                    lambda x: [x[0], x[1], x[6], x[1], x[5], x[1], x[4], x[1], x[3], x[1], x[2]],
                    # in sequence: Ar36, Ar37, Ar38, Ar39, Ar40
                    [value[i] for i in range(row_start_number + 2, row_stop_number - 7, 1)]))
            ]
            step_list.append(step_values)
    except Exception as e:
        raise ValueError('Error in opening the original file: %s' % str(e))
    else:
        return {'data': step_list}


def open_raw_txt(file_path, input_filter: List[Union[str, int]]):
    """
    Parameters
    ----------
    input_filter
    file_path

    Returns
    -------

    """
    if not input_filter:
        raise ValueError("Input filter is empty array.")

    if os.path.splitext(file_path)[1][1:].lower() != input_filter[0].strip().lower():
        raise ValueError("The file does not comply with the extension in the given filter.")

    with open(file_path, 'rb') as f:
        contents = f.read()
        encoding = chardet.detect(contents)
        lines = [line.strip().split(['\t', ';', " ", ",", input_filter[3]][int(input_filter[2])])
                 for line in contents.decode(encoding=encoding["encoding"]).split('\r\n')]

    file_name = os.path.basename(file_path).rstrip(os.path.splitext(file_path)[-1])
    step_list = get_raw_data([lines], input_filter, file_name=file_name)
    return {'data': step_list}


def open_raw_xls(file_path, input_filter: List[Union[str, int]]):
    """
    Parameters
    ----------
    file_path
    input_filter

    Returns
    -------

    """
    if not input_filter:
        raise ValueError("Input filter is empty array.")

    if os.path.splitext(file_path)[1][1:].lower() != input_filter[0].strip().lower():
        raise ValueError("The file does not comply with the extension in the given filter.")

    def _get_content_from_sheet(_index) -> List[List[Union[str, bool, int, float]]]:
        try:
            _sheet = wb.sheet_by_index(_index)
        except IndexError:
            return []
        else:
            return [[_sheet.cell(_row, _col).value for _col in range(_sheet.ncols)] for _row in range(_sheet.nrows)]

    wb = open_workbook(file_path)
    contents = [_get_content_from_sheet(i) for i in range(100)]
    file_name = os.path.basename(file_path).rstrip(os.path.splitext(file_path)[-1])
    step_list = get_raw_data(contents, input_filter, file_name=file_name)

    return {'data': step_list}


def open_raw_seq(file_path, input_filter=None):
    with open(file_path, 'rb') as f:
        sequences = pickle.load(f)
    name_list = []
    for seq in sequences:
        while seq.name in name_list:
            seq.name = f"{seq.name}-{seq.index}"
        name_list.append(seq.name)

    return {'sequences': sequences}


def get_raw_data(file_contents: List[List[Union[int, float, str, bool, list]]], input_filter: list,
                 file_name: str = "") -> list:
    """
    Parameters
    ----------
    file_name
    file_contents
    input_filter

    Returns
    -------

    """

    def datetime_parse(string, f):
        try:
            return datetime.strptime(string, f)
        except ValueError as v:
            if f.strip() == "":
                return datetime_parser.parse(string)
            elif len(v.args) > 0 and v.args[0].startswith('unconverted data remains: '):
                string = string[:-(len(v.args[0]) - 26)]
                return datetime.strptime(string, f)
            else:
                raise

    step_list = []
    idx = step_index = 0

    header = input_filter[5]
    isotopic_data_index = input_filter[8:28]
    sample_info_index = input_filter[33:65]
    optional_info_index = input_filter[37:-6]
    check_box_index = input_filter[-6:]

    timezone = sample_info_index[3] if sample_info_index[3] != "" else "utc"
    while True:  # measurment steps sloop
        # Zero datetime
        try:
            if check_box_index[2]:  # input_filter[134]: date in one string
                zero_date = datetime_parse(
                    get_item(file_contents, sample_info_index[16:19], base=[1, 1 - idx, 1]), sample_info_index[1])
            else:
                zero_date = datetime(year=get_item(file_contents, sample_info_index[16:19], base=1),
                                     month=get_item(file_contents, sample_info_index[22:25], base=[1, 1 - idx, 1]),
                                     day=get_item(file_contents, sample_info_index[28:31], base=[1, 1 - idx, 1]))

            if check_box_index[3]:  # input_filter[135]: time in one string
                zero_time = datetime_parse(
                    get_item(file_contents, sample_info_index[19:22], base=[1, 1 - idx, 1]), sample_info_index[2])
            else:
                zero_time = datetime(year=2020, month=12, day=31,
                                     hour=get_item(file_contents, sample_info_index[19:22], base=[1, 1 - idx, 1]),
                                     minute=get_item(file_contents, sample_info_index[25:28], base=[1, 1 - idx, 1]),
                                     second=get_item(file_contents, sample_info_index[31:34], base=[1, 1 - idx, 1]))

            zero_datetime = datetime(
                zero_date.year, zero_date.month, zero_date.day, zero_time.hour, zero_time.minute, zero_time.second)
            # adjust to UTC
            zero_datetime = utc_dt(zero_datetime, tz=timezone).isoformat(timespec='seconds')
        except (TypeError, ValueError, IndexError):
            # print(f"Cannot parse zero datetime")
            zero_datetime = datetime(1970, 1, 1, 0, 0, 0).isoformat(timespec='seconds')

        # Experiment name
        try:
            experiment_name = get_item(file_contents, sample_info_index[4:7], default="", base=[1, 1 - idx, 1])
        except (TypeError, ValueError, IndexError):
            # print(f"Cannot parse experiment name")
            experiment_name = "ExpName"

        # Step name
        try:
            step_name = get_item(file_contents, sample_info_index[7:10], default="", base=[1, 1 - idx, 1]) if input_filter[7] > 0 else ""
            if check_box_index[1] and sample_info_index[0].strip() != "":
                _res = string_parser(sample_info_index[0], file_name)
                if _res is not None:
                    experiment_name = _res.named.get("en", experiment_name)
                    step_index = _res.named.get("sn", step_name)
                    if step_index.isnumeric():
                        step_name = f"{experiment_name}-{int(step_index):02d}"
                    else:
                        step_name = f"{experiment_name}-{step_index}"
            if step_name == "":
                raise ValueError(f"Step name not found")
        except (TypeError, ValueError, IndexError):
            # When parsing the step name fails, the end of the file has been reached
            raise

        # other information
        options = get_sample_info(file_contents, optional_info_index, default="", base=[1, 1 - idx, 1])

        current_step = [[step_name, zero_datetime, experiment_name, options]]

        break_num = 0
        cycle_num = 0
        f = float(input_filter[31])  # Intensity Scale Factor
        data_content = file_contents[input_filter[4] - 1 if input_filter[4] != 0 else 0]
        for i in range(2000):  # measurement cycle sloop
            if break_num < input_filter[29]:
                break_num += 1
                continue
            break_num = 0
            if int(input_filter[6]) == 0:  # == 0, vertical
                start_row = input_filter[28] * cycle_num + input_filter[29] * cycle_num + header + idx - 1
                try:
                    current_step.append([
                        str(cycle_num + 1),
                        # in sequence: Ar36, Ar37, Ar38, Ar39, Ar40
                        float(data_content[start_row + isotopic_data_index[18]][isotopic_data_index[19] - 1]),
                        float(data_content[start_row + isotopic_data_index[16]][isotopic_data_index[17] - 1]) * f,
                        float(data_content[start_row + isotopic_data_index[14]][isotopic_data_index[15] - 1]),
                        float(data_content[start_row + isotopic_data_index[12]][isotopic_data_index[13] - 1]) * f,
                        float(data_content[start_row + isotopic_data_index[10]][isotopic_data_index[11] - 1]),
                        float(data_content[start_row + isotopic_data_index[ 8]][isotopic_data_index[ 9] - 1]) * f,
                        float(data_content[start_row + isotopic_data_index[ 6]][isotopic_data_index[ 7] - 1]),
                        float(data_content[start_row + isotopic_data_index[ 4]][isotopic_data_index[ 5] - 1]) * f,
                        float(data_content[start_row + isotopic_data_index[ 2]][isotopic_data_index[ 3] - 1]),
                        float(data_content[start_row + isotopic_data_index[ 0]][isotopic_data_index[ 1] - 1]) * f,
                    ])
                except (ValueError, IndexError):
                    print(f"Cannot parse isotope data")
                    print(traceback.format_exc())
                    current_step.append([
                        str(cycle_num + 1), None, None, None, None, None, None, None, None, None, None,
                    ])
            elif int(input_filter[6]) == 1:  # == 1, horizontal
                start_row = input_filter[5] + idx
                col_inc = input_filter[28] * cycle_num + input_filter[29] * cycle_num - 1
                try:
                    current_step.append([
                        str(cycle_num + 1),
                        # Ar36, Ar37, Ar38, Ar39, Ar40
                        float(data_content[start_row][isotopic_data_index[19] + col_inc]),
                        float(data_content[start_row][isotopic_data_index[17] + col_inc]) * f,
                        float(data_content[start_row][isotopic_data_index[15] + col_inc]),
                        float(data_content[start_row][isotopic_data_index[13] + col_inc]) * f,
                        float(data_content[start_row][isotopic_data_index[11] + col_inc]),
                        float(data_content[start_row][isotopic_data_index[ 9] + col_inc]) * f,
                        float(data_content[start_row][isotopic_data_index[ 7] + col_inc]),
                        float(data_content[start_row][isotopic_data_index[ 5] + col_inc]) * f,
                        float(data_content[start_row][isotopic_data_index[ 3] + col_inc]),
                        float(data_content[start_row][isotopic_data_index[ 1] + col_inc]) * f,
                    ])
                except (ValueError, IndexError):
                    print(f"Cannot parse isotope data")
                    print(traceback.format_exc())
                    current_step.append([
                        str(cycle_num + 1), None, None, None, None, None, None, None, None, None, None,
                    ])
            else:
                raise ValueError(f"{input_filter[6]} not in [0, 1]")

            cycle_num += 1
            if cycle_num >= input_filter[7]:
                break

        step_list.append(current_step)
        idx = input_filter[32] * len(step_list)

        if not check_box_index[0] or len(step_list) >= 500:  # check_box_index[0]: multiple sequences
            break

    return step_list


def get_sample_info(file_contents: list, index_list: list, default="", base: Union[int, tuple, list] = 1) -> dict:
    """
    Parameters
    ----------
    file_contents
    index_list
    default
    base

    Returns
    -------

    """
    sample_info = DEFAULT_SAMPLE_INFO.copy()
    sample_info.update({
        "ExpName": get_item(file_contents, index_list[0:3], default=default, base=base),
        "StepName": get_item(file_contents, index_list[3:6], default=default, base=base),
        "SmpType": get_item(file_contents, index_list[6:9], default=default, base=base),
        "StepLabel": get_item(file_contents, index_list[9:12], default=default, base=base),
        "ZeroYear": get_item(file_contents, index_list[12:15], default=default, base=base),  # year
        "ZeroHour": get_item(file_contents, index_list[15:18], default=default, base=base),  # hour
        "ZeroMon": get_item(file_contents, index_list[18:21], default=default, base=base),  # month
        "ZeroMin": get_item(file_contents, index_list[21:24], default=default, base=base),  # minute
        "ZeroDay": get_item(file_contents, index_list[24:27], default=default, base=base),  # day
        "ZeroSec": get_item(file_contents, index_list[27:30], default=default, base=base),  # second
        "SmpName": get_item(file_contents, index_list[30:33], default=default, base=base),
        "SmpLoc": get_item(file_contents, index_list[33:36], default=default, base=base),
        "SmpMatr": get_item(file_contents, index_list[36:39], default=default, base=base),
        "ExpType": get_item(file_contents, index_list[39:42], default=default, base=base),
        "SmpWeight": get_item(file_contents, index_list[42:45], default=default, base=base),
        "Stepunit": get_item(file_contents, index_list[45:48], default=default, base=base),
        "HeatingTime": get_item(file_contents, index_list[48:51], default=default, base=base),
        "InstrName": get_item(file_contents, index_list[51:54], default=default, base=base),
        "Researcher": get_item(file_contents, index_list[54:57], default=default, base=base),
        "Analyst": get_item(file_contents, index_list[57:60], default=default, base=base),
        "Lab": get_item(file_contents, index_list[60:63], default=default, base=base),
        "Jv": get_item(file_contents, index_list[63:66], default=default, base=base),
        "Jsig": get_item(file_contents, index_list[66:69], default=default, base=base),
        "CalcName": get_item(file_contents, index_list[69:72], default=default, base=base),
        "IrraName": get_item(file_contents, index_list[72:75], default=default, base=base),
        "IrraLabel": get_item(file_contents, index_list[75:78], default=default, base=base),
        "IrraPosH": get_item(file_contents, index_list[78:81], default=default, base=base),
        "IrraPosX": get_item(file_contents, index_list[81:84], default=default, base=base),
        "IrraPosY": get_item(file_contents, index_list[84:87], default=default, base=base),
        "StdName": get_item(file_contents, index_list[87:90], default=default, base=base),
        "StdAge": get_item(file_contents, index_list[90:93], default=default, base=base),
        "StdAgeSig": get_item(file_contents, index_list[93:96], default=default, base=base),
        # "Experiment Name": get_item(file_contents, sample_info_index[0:3], default=default, base=base),
        # "Step Name": get_item(file_contents, sample_info_index[3:6], default=default, base=base),
        # "Sample Type": get_item(file_contents, sample_info_index[6:9], default=default, base=base),
        # "Step Label": get_item(file_contents, sample_info_index[9:12], default=default, base=base),
        # "Zero Date Year": get_item(file_contents, sample_info_index[12:15], default=default, base=base),
        # "Zero Time Hour": get_item(file_contents, sample_info_index[15:18], default=default, base=base),
        # "Zero Date Month": get_item(file_contents, sample_info_index[18:21], default=default, base=base),
        # "Zero Time Minute": get_item(file_contents, sample_info_index[21:24], default=default, base=base),
        # "Zero Date Day": get_item(file_contents, sample_info_index[24:27], default=default, base=base),
        # "Zero Time Second": get_item(file_contents, sample_info_index[27:30], default=default, base=base),
        # "Sample Name": get_item(file_contents, sample_info_index[30:33], default=default, base=base),
        # "Sample location": get_item(file_contents, sample_info_index[33:36], default=default, base=base),
        # "Sample Material": get_item(file_contents, sample_info_index[36:39], default=default, base=base),
        # "Experiment Type": get_item(file_contents, sample_info_index[39:42], default=default, base=base),
        # "Sample Weight": get_item(file_contents, sample_info_index[42:45], default=default, base=base),
        # "Step unit": get_item(file_contents, sample_info_index[45:48], default=default, base=base),
        # "Heating time": get_item(file_contents, sample_info_index[48:51], default=default, base=base),
        # "Instrument name": get_item(file_contents, sample_info_index[51:54], default=default, base=base),
        # "Researcher": get_item(file_contents, sample_info_index[54:57], default=default, base=base),
        # "Analyst": get_item(file_contents, sample_info_index[57:60], default=default, base=base),
        # "Laboratory": get_item(file_contents, sample_info_index[60:63], default=default, base=base),
        # "J value": get_item(file_contents, sample_info_index[63:66], default=default, base=base),
        # "J value error": get_item(file_contents, sample_info_index[66:69], default=default, base=base),
        # "Calc params": get_item(file_contents, sample_info_index[69:72], default=default, base=base),
        # "Irra name": get_item(file_contents, sample_info_index[72:75], default=default, base=base),
        # "Irra label": get_item(file_contents, sample_info_index[75:78], default=default, base=base),
        # "Irra position H": get_item(file_contents, sample_info_index[78:81], default=default, base=base),
        # "Irra position X": get_item(file_contents, sample_info_index[81:84], default=default, base=base),
        # "Irra position Y": get_item(file_contents, sample_info_index[84:87], default=default, base=base),
        # "Standard name": get_item(file_contents, sample_info_index[87:90], default=default, base=base),
        # "Standard age": get_item(file_contents, sample_info_index[90:93], default=default, base=base),
        # "Standard age error": get_item(file_contents, sample_info_index[93:96], default=default, base=base),
    })
    return sample_info



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - calc - jvalue
# ==========================================
#
#
#
"""
import numpy as np


def j_value(age, sage, r, sr, f, rsf):
    """ Calculate J value according to the given age and the ratio

    Parameters
    ----------
    age : age of the reference standard sample, in Ma
    sage : 1 sigma error of age
    r : 40/39 ratio of the reference standard standard sample
    sr : 1 sigma error of 40/39 ratio
    f : decay constant(lambda) of K
    rsf : relative error of decay constant

    Returns
    -------
    tuple of  J value and error
    """
    age, sage, r, sr, f, rsf = np.array([age, sage, r, sr, f, rsf])
    f = f * 1000000  # exchange to unit of Ma
    rsf = f * rsf / 100  # exchange to absolute error
    k0 = (np.exp(f * age) - 1) / r
    v1 = rsf ** 2 * (age * np.exp(f * age) / r) ** 2
    v2 = sage ** 2 * (f * np.exp(f * age) / r) ** 2
    v3 = sr ** 2 * ((1 - np.exp(f * age)) / r ** 2) ** 2
    k1 = np.sqrt(v1 + v2 + v3)
    return k0, k1



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - calc - raw_funcs
# ==========================================
#
#
#
"""
import traceback
import numpy as np
from .arr import transpose, is_twoD
from . import regression


"""get regression results for raw data points"""


def get_raw_data_regression_results(points_data, unselected: list = None):
    """
    Parameters
    ----------
    points_data : two dimensional list. like [[x1, y1], [x2, y2], ..., [xn, yn]]
    unselected :

    Returns
    -------

    """
    # if unselected is None:
    #     unselected = []
    # linesData = []
    linesResults, regCoeffs = [], []
    # un_x = transpose(unselected)[0] if is_twoD(unselected) else []
    reg_handler = [
        regression.linest, regression.quadratic, regression.exponential,
        regression.power, regression.average]
    # size = 50
    # lines_x = [(max(x + un_x) - 0) / size * i for i in range(size + 1)]
    for i in range(len(reg_handler)):
        try:
            x, y = transpose(points_data, ignore=False)
            res = reg_handler[i](a0=y, a1=x)
            # line_data = transpose([lines_x, res[7](lines_x)])
            line_results = res[0:4]
            reg_coeffs = res[5]
            if any(np.isnan(line_results)):
                raise ValueError
            # if np.isin(np.inf, line_data) or np.isin(np.nan, line_data):
            #     raise ZeroDivisionError(f"Infinite value or nan value.")
            if abs(res[0] - min(y)) > 5 * (max(y) - min(y)):
                raise ValueError
        except RuntimeError:
            line_data, line_results, reg_coeffs = [], ['RuntimeError', np.nan, np.nan, np.nan, ], []
        except np.linalg.LinAlgError:
            line_data, line_results, reg_coeffs = [], ['MatrixError', np.nan, np.nan, np.nan, ], []
        except TypeError or IndexError:
            line_data, line_results, reg_coeffs = [], ['NotEnoughPoints', np.nan, np.nan, np.nan, ], []
        except ZeroDivisionError:
            line_data, line_results, reg_coeffs = [], ['ZeroDivisionError', np.nan, np.nan, np.nan, ], []
        except ValueError:
            line_data, line_results, reg_coeffs = [], ['BadFitting', np.nan, np.nan, np.nan, ], []
        except:
            line_data, line_results, reg_coeffs = [], ['UncaughtError', np.nan, np.nan, np.nan, ], []
        # linesData.append(line_data)
        linesResults.append(line_results)
        regCoeffs.append(reg_coeffs)
    return None, linesResults, regCoeffs




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - files - basic
# ==========================================
#
#
#
"""

import os
import pickle
import traceback
import json


def upload(file, media_dir):
    try:
        name, suffix = os.path.splitext(file.name)
        if suffix.lower() not in [
            '.xls', '.age', '.xlsx', '.arr', '.jpg', '.png', '.txt',
            '.log', '.seq', '.json', '.ahd', '.csv']:
            raise TypeError(f"The imported file is not supported: {suffix}")
        web_file_path = os.path.join(media_dir, file.name)
        with open(web_file_path, 'wb') as f:
            for chunk in file.chunks():
                f.write(chunk)
        print("File path on the server: %s" % web_file_path)
    except PermissionError:
        raise ValueError(f'Permission denied')
    except (Exception, BaseException) as e:
        print(traceback.format_exc())
        raise ValueError(f'Error in opening file: {e}')
    else:
        return web_file_path, name, suffix


def read(file_path):
    """ Read text files, default 'r', 'rb'
    Parameters
    ----------
    file_path

    Returns
    -------

    """
    try:
        with open(file_path, 'r') as f:
            params = json.load(f)
    except UnicodeDecodeError:
        with open(file_path, 'rb') as f:
            params = pickle.load(f)
    return params


def write(file_path, params):
    """
    Parameters
    ----------
    file_path
    params

    Returns
    -------

    """
    # with open(file_path, 'wb') as f:
    #     f.write(pickle.dumps(params))
    with open(file_path, 'w') as f:  # save serialized json data to a readable text
        f.write(json.dumps(params))
    return file_path


def delete(file_path):
    """
    Parameters
    ----------
    file_path

    Returns
    -------

    """
    try:
        os.remove(file_path)
    except Exception:
        return False
    else:
        return True



#  Copyright (C) 2024 Yang. - All Rights Reserved

#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2024 Yang 
# ararpy - test_regression_methods
# ==========================================
#
#
# 
"""

import ararpy as ap
import numpy as np
from math import exp
from scipy.optimize import minimize
from matplotlib import pyplot as plt
import matplotlib
matplotlib.use('TkAgg')


# def get_rand_measurment():
#     geological_error = 0.0000001  # relative
#     analytical_error = 0.01  # relative
#     J = 0.01
#     L = 5.53E-10
#     I = 1 / 298.56
#
#     true_age = 100 * 1000000
#     true_F = (exp(true_age * L) - 1) / J
#     size = 10
#     ages = np.random.normal(loc=true_age, scale=geological_error * true_age, size=size)
#     Fs = [(exp(age * L) - 1) / J for age in ages]
#     F = [np.random.normal(loc=(exp(ages[i] * L) - 1) / J, scale=(exp(ages[i] * L) - 1) / J * analytical_error, size=1)[0] for i in range(size)]
#
#     ar36a = np.array([0.1 + 0.1 * i for i in range(size)])
#     sar36a = ar36a * analytical_error
#
#     ar40a = ar36a / I
#     sar40a = sar36a / I
#
#     ar40 = np.array([350 + 10 * i for i in range(size)])
#     sar40 = ar40 * analytical_error
#
#     ar40r = ar40 - ar40a
#     sar40r = (sar40 ** 2 + sar40a ** 2) ** .5
#     ar39k = [ar40r[i] / F[i] for i in range(size)]
#     sar39k = [sar40r[i] / F[i] for i in range(size)]
#
#     [x, sx] = ap.calc.arr.div([ar39k, sar39k], [ar40.tolist(), sar40.tolist()])  # ar39k / ar40
#     [y, sy] = ap.calc.arr.div([ar36a.tolist(), sar36a.tolist()], [ar40.tolist(), sar40.tolist()])  # ar36a / ar40
#
#     r = [ap.calc.err.cor(sar39k[i] / ar39k[i], sar36a[i] / ar36a[i], sar40[i] / ar40[i]) for i in range(size)]
#
#     print(f"{r = }")
#     print(f"{F = }")
#     print(f"{I = }")
#
#     with open(f"test_regression_methods.text", "w") as f:
#         f.write(f"{geological_error = }\n{analytical_error = }\n{J = }\n{L = }\n{I = }\n{true_age = }\n{true_F = }\n{size = }\n")
#         f.write(f"age,F,ar40,s,ar36a,s,ar39k,s,x,sx,y,sy,r\n")
#         f.writelines([f"{ages[i]},{F[i]},{ar40[i]},{sar40[i]},{ar36a[i]},{sar36a[i]},{ar39k[i]},{sar39k[i]},{x[i]},{sx[i]},{y[i]},{sy[i]},{r[i]}\n" for i in range(size)])
#
#     return x, sx, y, sy, r


size = 5
geological_error = [0.001 for i in range(size)]  # relative
analytical_error = [0.001 for i in range(size)]  # relative
geological_error[2] = 0.003  # relative
analytical_error[2] = 0.003  # relative
J = 0.01
L = 5.53E-10
I = 298.56

true_age = 100 * 1000000
true_F = (exp(true_age * L) - 1) / J


def get_rand_measurement(normal: bool = True):

    ages = [np.random.normal(loc=true_age, scale=geological_error[i] * true_age, size=1)[0] for i in range(size)]
    # ages[2] = 110000000
    F = [(exp(age * L) - 1) / J for age in ages]
    # F = [np.random.normal(loc=(exp(ages[i] * L) - 1) / J, scale=(exp(ages[i] * L) - 1) / J * analytical_error, size=1)[0] for i in range(size)]

    ar39k = [300 + i * 50 for i in range(size)]
    sar39k = [ar39k[i] * analytical_error[i] for i in range(size)]

    ar36a = sorted([10 + i * 0.5 for i in range(size)], reverse=True)
    sar36a = [ar36a[i] * analytical_error[i] for i in range(size)]

    ar40 = [ar36a[i] * I + ar39k[i] * F[i] for i in range(size)]
    sar40 = [ap.calc.err.add(I * sar36a[i], F[i] * sar39k[i]) for i in range(size)]

    ar36a = [np.random.normal(loc=val, scale=sar36a[i], size=1)[0] for i, val in enumerate(ar36a)]
    ar39k = [np.random.normal(loc=val, scale=sar39k[i], size=1)[0] for i, val in enumerate(ar39k)]
    ar40 = [np.random.normal(loc=val, scale=sar40[i], size=1)[0] for i, val in enumerate(ar40)]

    if normal:
        [x, sx] = ap.calc.arr.div([ar39k, sar39k], [ar36a, sar36a])  # ar39k / ar40
        [y, sy] = ap.calc.arr.div([ar40, sar40], [ar36a, sar36a])  # ar36a / ar40
        r = [ap.calc.err.cor(sar39k[i] / ar39k[i], sar40[i] / ar40[i], sar36a[i] / ar36a[i]) for i in range(size)]

    else:
        [x, sx] = ap.calc.arr.div([ar36a, sar36a], [ar40, sar40])  # ar39k / ar40
        [y, sy] = ap.calc.arr.div([ar39k, sar39k], [ar40, sar40])  # ar36a / ar40
        r = [ap.calc.err.cor(sar36a[i] / ar36a[i], sar39k[i] / ar39k[i], sar40[i] / ar40[i]) for i in range(size)]
        F = [1/i for i in F]

    # [x, sx] = ap.calc.arr.div([ar39k, sar39k], [ar40, sar40])  # ar39k / ar40
    # [y, sy] = ap.calc.arr.div([ar36a, sar36a], [ar40, sar40])  # ar36a / ar40
    # r = [ap.calc.err.cor(sar39k[i] / ar39k[i], sar36a[i] / ar36a[i], sar40[i] / ar40[i]) for i in range(size)]

    print(f"{ages = }")
    print(f"{r = }")
    print(f"{F = }")
    print(f"{I = }")
    print(f"{true_F = }\n")

    with open(f"test_regression_methods.text", "w") as f:
        f.write(f"{geological_error = }\n{analytical_error = }\n{J = }\n{L = }\n{I = }\n{true_age = }\n{true_F = }\n{size = }\n")
        f.write(f"age,F,ar40,s,ar36a,s,ar39k,s,x,sx,y,sy,r\n")
        f.writelines([f"{ages[i]},{F[i]},{ar40[i]},{sar40[i]},{ar36a[i]},{sar36a[i]},{ar39k[i]},{sar39k[i]},{x[i]},{sx[i]},{y[i]},{sy[i]},{r[i]}\n" for i in range(size)])

    return x, sx, y, sy, r, F


# 定义线性模型
def linear_model(params, x):
    slope, intercept = params
    return slope * x + intercept


# 定义负对数似然函数
def negative_log_likelihood(params, x, y):
    predicted_y = linear_model(params, x)
    # 使用高斯误差进行似然计算
    errors = predicted_y - y
    likelihood = np.sum(-0.5 * np.log(2 * np.pi) - 0.5 * errors**2)
    return -likelihood


def do_repetition():
    york2 = []
    linest = []
    mle = []
    Fs = []
    data = [[], [], [], []]
    for i in range(100):
        [x, sx, y, sy, r, F] = get_rand_measurement(False)

        for j in range(len(x)):
            data[0].append(x[j])
            data[1].append(sx[j])
            data[2].append(y[j])
            data[3].append(sy[j])

        ap.calc.arr.multi_append(Fs, np.mean(F))

        res = ap.calc.regression.york2(x, sx, y, sy, r)
        print(f"Intercept = {res[0]} ± {res[1]}")
        print(f"Slope = {res[2]} ± {res[3]}")
        print(f"MSWD = {res[4]}")
        print(f"diff = {true_F - res[2]}\n")

        york2.append(res[:5])

        res = ap.calc.regression.linest(y, x)
        print(f"Intercept = {res[5][0]} ± {res[6][0]}")
        print(f"Slope = {res[5][1]} ± {res[6][1]}")
        print(f"MSWD = {res[4]}")
        print(f"diff = {true_F - res[5][1]}\n")

        linest.append([res[5][0], res[6][0], res[5][1], res[6][1], res[4]])

        res = minimize(negative_log_likelihood, [res[2], res[0]], args=(np.array(x), np.array(y)))
        print(res.items())
        slope, intercept = res.x
        mle.append([intercept, 0, slope, 0, "MSWD"])

    return york2, linest, mle, Fs, data


fig, ((ax1, ax2), (ax3, ax4), (ax5, ax6)) = plt.subplots(3, 2)

york2, linest, mle, Fs, data = do_repetition()

# for each in york2:
#     ax1.plot((each[0] - each[1], each[0] + each[1]), (each[2], each[2]), '-')

york2, linest, mle = ap.calc.arr.transpose(york2), ap.calc.arr.transpose(linest), ap.calc.arr.transpose(mle)

ax1.scatter(york2[0], york2[2], c="black", s=3)
ax1.scatter(Fs, [np.median(york2[2]) for i in Fs], c="blue", s=1)
ax2.scatter(linest[0], linest[2], c="red", s=3)
ax2.scatter(Fs, [np.median(linest[2]) for i in Fs], c="blue", s=1)
# ax3.scatter(mle[0], mle[2], c="blue", s=3)
# ax3.scatter(Fs, [np.median(mle[2]) for i in Fs], c="blue", s=1)

# ax3.set_ylim(0.150, 0.220)
# ax4.set_ylim(0.150, 0.220)

ax3.set_ylim(0.160, 0.190)
ax4.set_ylim(0.160, 0.190)

ax3_2 = ax3.twinx()

for i in range(len(Fs)):
    ax3.plot([i, i], [york2[0][i]-york2[1][i], york2[0][i]+york2[1][i]], '-', color='black')
    ax3.scatter([i], [Fs[i]], c="red", s=1)

ax3_2.plot([i for i in range(len(Fs))], york2[4], '-', color="blue")


for i in range(len(york2[0])):
    ax5.plot([0.00165, 0.0023], [j * york2[2][i] + york2[0][i] for j in [0.00165, 0.0023]], color='red')

ax5.scatter(data[0], data[2], c="blue", s=2)


for i in range(len(Fs)):
    ax4.plot([i, i], [linest[0][i]-linest[1][i], linest[0][i]+linest[1][i]], '-', color='black')
    ax4.scatter([i], [Fs[i]], c="red", s=1)


for i in range(len(york2[0])):
    ax6.plot([0.00165, 0.0023], [j * linest[2][i] + linest[0][i] for j in [0.00165, 0.0023]], color='red')

ax6.scatter(data[0], data[2], c="blue", s=2)

#
# [x, sx, y, sy, r] = get_rand_measurement()
#
# # for i in range(len(x)):
# #     cov = [[sx[i] ** 2, r[i] * sx[i] * sy[i]], [r[i] * sx[i] * sy[i], sy[i] ** 2]]
# #     _x, _y = np.random.multivariate_normal([x[i], y[i]], cov, 4000).transpose()
# #     plt.scatter(_x, _y, c="b", s=1)
# #
# #
# # plt.scatter(x, y, c="black", s=1)
#
# line_x = np.linspace(20, 60, 50)
#
# res = ap.calc.regression.york2(x, sx, y, sy, r)
# print(f"Intercept = {res[0]} ± {res[1]}")
# print(f"Slope = {res[2]} ± {res[3]}")
# print(f"MSWD = {res[894]}")
# print(f"diff = {true_F - res[2]}\n")
# # plt.plot(line_x, line_x * res[2] + res[0])
#
#
# res = ap.calc.regression.linest(y, x)
# print(f"Intercept = {res[5][0]} ± {res[6][0]}")
# print(f"Slope = {res[5][1]} ± {res[6][1]}")
# print(f"MSWD = {res[4]}")
# print(f"diff = {true_F - res[5][1]}\n")
# plt.plot(line_x, line_x * res[5][1] + res[5][0])


# res = ap.calc.regression.monte_carlo_linest(x, sx, y, sy, r, iteration=1000)
# slope = [i[1] for i in res]
# intercept = [i[0] for i in res]
# cov = [i[2] for i in res]
# print(f"Intercept = {np.mean(intercept)} ± {np.std(intercept)}")
# print(f"Slope = {np.mean(slope)} ± {np.std(slope)}")
# print(f"diff = {true_F - np.mean(slope)}\n")
# plt.plot(line_x, line_x * np.mean(slope) + np.mean(intercept))
#
# k1, k2 = [], []
# for i, val in enumerate(cov):
#     _intercept, _slope = np.random.multivariate_normal([intercept[i], slope[i]], val, 500).transpose()
#     k1 = [*k1, *_intercept]
#     k2 = [*k2, *_slope]
#     ax6.scatter(_slope, _intercept, c="grey")
# _slope = np.mean(slope)
# _intercept = np.mean(intercept)

# print(f"Intercept = {np.mean(k1)} ± {np.std(k1)}")
# print(f"Slope = {np.mean(k2)} ± {np.std(k2)}")
# print(f"diff = {true_F - np.mean(k2)}\n")
# plt.plot(line_x, line_x * np.mean(k2) + np.mean(k1))


#
# e = 0.1
# x = [100, 150, 200, 250, 300]
# sx = [_x * e for _x in x]
# y = [100, 125, 150, 175, 190]
# sy = [_y * e for _y in y]
# r = [0.9, 0.9, 0.9, 0.9, 0.9]
#
#
# res = ap.calc.regression.monte_carlo_linest(x, sx, y, sy, r, iteration=1000)
#
# slope = [i[1] for i in res]
# intercept = [i[0] for i in res]
# cov = [i[2] for i in res]
#
# # for i, val in enumerate(cov):
# #     _intercept, _slope = np.random.multivariate_normal([intercept[i], slope[i]], val, 500).transpose()
# #     plt.scatter(_slope, _intercept, c="grey")
#
# # plt.scatter(slope, intercept, c="blue")
#
# res = ap.calc.regression.linest(y, x)
# intercept, slope = res[5]
# intercepts, slopes = np.random.multivariate_normal([intercept, slope], res[-1], 1000).transpose()
# plt.scatter(slopes, intercepts, c="orange")
# plt.scatter(slope, intercept, c="red")
#
# res = ap.calc.regression.york2(x, sx, y, sy, r)
# plt.scatter(res[2], res[0], c="black")


# 显示图形
plt.show()





#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - files - arr_file
# ==========================================
#
#
#
"""
# === External imports ===
import os
import pickle


def save(file_path, sample):
    """ Save arr project as arr files

    Parameters
    ----------
    file_path : str, filepath
    sample : Sample instance

    Returns
    -------
    str, file name
    """
    file_path = os.path.join(file_path, f"{sample.Info.sample.name}.arr")
    with open(file_path, 'wb') as f:
        f.write(pickle.dumps(sample))
    # with open(file_path, 'w') as f:
    # # save serialized json data to a readable text
    #     f.write(basic_funcs.getJsonDumps(sample))
    return f"{sample.Info.sample.name}.arr"





#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2024 Yang
# argon atomic_level_random_walk model - main
# ==========================================
#
#
#
"""
import os
import pickle
import numpy as np
import math
import copy
import time
from typing import List

np.set_printoptions(precision=18, threshold=10000, linewidth=np.inf)

SETUP_PRINT = "natoms = {0}, nsteps = {1}, grain_size = {2}, time_scale = {3}, size_scale = {4}, number_scale = {5}"


def _uniform_sphere_step(step_length, shape):
    """
    在三维空间中生成均匀分布的球面样本，且步长严格为指定长度。

    Parameters:
        step_length (array, float): 步长的固定长度, (n, m)
        num_samples (tuple): 步长shape, (n, m)

    Returns:
        np.ndarray: 形状为 (num_samples, 3) 的数组，每行是一个随机步长向量
    """
    # 随机生成方位角 phi 和极角的 cos(theta)
    num_samples, dimension = shape
    phi = np.random.uniform(0, 2 * np.pi, num_samples)
    cos_theta = np.random.uniform(-1, 1, num_samples)
    sin_theta = np.sqrt(1 - cos_theta ** 2)  # sin(theta) = sqrt(1 - cos^2(theta))

    # 将球面坐标转换为笛卡尔坐标
    x = sin_theta * np.cos(phi)
    y = sin_theta * np.sin(phi)
    z = cos_theta

    # 构建单位向量
    unit_vectors = np.vstack((x, y, z)).T

    # 乘以步长，得到最终的步长向量
    steps = step_length * unit_vectors
    return steps


def walker(pos, step_length, total_nsteps, min_bound, max_bound, scale=0., compensation=0, remove=True,
           conditions=None, boundary_factor=1, decay=0, parent=0):

    if len(pos) == 0:
        return pos

    dimension = pos.shape[1] if len(pos.shape) > 1 else 1

    sigma = step_length * math.sqrt(scale)

    if conditions is None:
        if dimension == 3:
            conditions = np.array([[-50, -50, -50, 50, 50, 50, 1]])
        else:
            conditions = np.array([[-50, -50, 50, 50, 1]])

    num_big_steps = int(total_nsteps // scale) if int(scale) > 1 else 0
    num_small_steps = int(total_nsteps % scale) if int(scale) > 1 else int(total_nsteps)

    for _ in range(num_big_steps):
        # print(f"new walker {_ = } {len(pos) = }")
        # 边界判断（布尔掩码）
        in_boundary_mask = np.all(
            (pos >= (min_bound + sigma * compensation)) & (pos <= (max_bound - sigma * compensation)), axis=1)
        in_core = pos[in_boundary_mask]
        in_boundary = pos[~in_boundary_mask]

        # 核心粒子随机行走
        coefficients = np.ones(len(in_core))
        for each in conditions:
            # each = x1, y1, z1, x2, y2, z2, coeff for dimension=3
            cond = np.all(
                (in_core >= np.array(each[:dimension])) & (in_core <= np.array(each[dimension:int(2*dimension)])),
                axis=1)
            coefficients = np.where(cond, each[-1] ** 0.5, coefficients)
        steps = np.random.normal(0, sigma, size=in_core.shape)
        in_core += steps * coefficients[:, None]

        # 将边界粒子以更小 scale 的步长模拟
        if scale > 1:
            in_boundary = walker(in_boundary, step_length, scale, min_bound, max_bound, scale // 2,
                                 compensation, remove, conditions, boundary_factor)

        # 拼接核心和边界粒子
        pos = np.concatenate((in_boundary, in_core))

        # 移除超出范围的粒子
        if remove:
            pos = pos[np.all((pos >= min_bound) & (pos <= max_bound), axis=1)]

    # 最后处理 scale < 1 的小步模拟, boundary_factor用来缩放这里的处理步数，1 < boundary_factor <= 1如果 boundary_factor=1，将完全吻合实际
    # boundary_factor < 1 将加快一些速度， boundary_factor=0 将忽略小步数
    k = boundary_factor * max(conditions[:, -1])
    for _ in range(int(num_small_steps * k)):
        coefficients = np.ones(len(pos))
        for each in conditions:
            cond = np.all((pos >= np.array(each[:dimension])) & (pos <= np.array(each[dimension:int(2*dimension)])),
                          axis=1)
            coefficients = np.where(cond, (each[-1] / k) ** 0.5, coefficients)
            # coefficients = np.where(cond, coeff ** 0.5, coefficients)
        steps = np.random.normal(0, step_length, size=np.shape(pos))
        pos += steps * coefficients[:, None]

        if remove:
            pos = pos[np.all((pos >= min_bound) & (pos <= max_bound), axis=1)]

    return pos


def walker2(pos: np.ndarray, duration, step_length, min_bound, max_bound, time_scale, frequency,
            conditions=None, remove: bool = True, decay=0, parent=0):
    """
    :return:
    """

    if len(pos) == 0:
        return pos

    dimension = pos.shape[-1] if len(pos.shape) > 1 else 1


    dt = time_scale
    if conditions is None:
        if dimension == 3:
            conditions = np.array([[-50, -50, -50, 50, 50, 50, 1]])
        else:
            conditions = np.array([[-50, -50, 50, 50, 1]])

    for step in range(int(1e16)):

        n = parent - int(parent * math.exp(-1 * decay * dt / (3600 * 24 * 365.2425)))
        # print(f"number of new pos = {n}")
        parent = parent - n
        new_pos = np.random.uniform(max(max_bound), min(min_bound), size=(int(n), dimension))
        new_pos = new_pos[np.all((new_pos >= min_bound) & (new_pos <= max_bound), axis=1)]
        pos = np.concatenate((pos, new_pos), axis=0)

        res = np.zeros(0).reshape(0, dimension)
        for index, each in enumerate(conditions):
            # each = (x1, y1, z1, x2, y2, z2, gamma)
            d = step_length * math.sqrt(each[-1] * frequency * time_scale)

            locs = np.ones(len(pos)) * 999
            for _index, _each in enumerate(conditions):
                cond = np.all((pos >= np.array(_each[:dimension])) & (pos <= np.array(_each[dimension:int(2*dimension)])), axis=1)
                locs = np.where(cond, _index, locs)

            partial_pos = pos[locs == index]

            # c = len(partial_pos)
            # # 初始化步长
            # steps = np.zeros(partial_pos.shape, dtype=float)
            # # 移动方向, x or y or z
            # directions = np.random.choice(list(range(dimension)), size=c)
            # # 方向, positive or negative, 步长
            # steps[np.arange(c), directions] = np.random.choice([-d, d], size=c, p=[0.5, 0.5])
            steps = np.random.normal(0, d, size=partial_pos.shape)

            partial_pos += steps

            if remove:
                partial_pos = partial_pos[np.all((partial_pos >= min_bound) & (partial_pos <= max_bound), axis=1)]

            res = np.concatenate((res, partial_pos))

        pos = copy.deepcopy(res)

        _t = (step + 1) * dt

        if abs(_t - duration) < dt:
            break

    return pos


class OverEpsilonError(Exception):
    def __init__(self, i, d, e):
        self.i = i
        self.d = d
        self.e = e

    def __str__(self):
        return f"OverEpsilonError: Index = {self.i}, d = {self.d} > e = {self.e}"


class Domain:
    def __init__(self, dimension: int = 3, atom_density: float = 1e14, fraction: float = 1.0,
                 min_bound=None, max_bound=None, **kwargs):

        self.dimension = dimension
        self.size = ...
        self.atom_density = atom_density
        self.fraction = fraction
        self.time_scale = 1
        self.size_scale = 0.05

        self.energy = 120 * 1000  # J/mol
        self.density = 0
        self.fracture = 0
        self.boundary = 0
        self.defect = 0
        self.pho = 1
        self.step_length = 0.0001  # 1 A = 0.1 nm = 0.0001 μm

        self.current_temp = 273.15

        self.inclusions: List[Domain] = []
        self.natoms: int = 0
        self.positions = []
        self.min_bound = min_bound
        self.max_bound = max_bound
        self.remained_per_step = []
        self.released_per_step = []

        self.attr = {}

        for key, val in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, val)

        self.setup()

    def setup(self):
        """
        Used to set or reset positions of atoms
        :return:
        """

        # 初始化粒子位置，随机分布在网格内
        # 初始原子个数  1e14 原子密度 个/立方厘米   1e-4是单位换算 cm/µm
        natoms = int(self.atom_density ** (self.dimension / 3) * (abs(max(self.max_bound) - min(self.min_bound)) * 1e-4) ** self.dimension)
        natoms = 1 if natoms == 0 else natoms
        self.positions = np.random.uniform(max(self.max_bound), min(self.min_bound), size=(int(natoms), self.dimension))
        self.positions = self.positions[np.all((self.positions >= self.min_bound) & (self.positions <= self.max_bound), axis=1)]
        for dom in self.inclusions:
            self.positions = self.positions[~np.all((self.positions >= dom.min_bound) & (self.positions <= dom.max_bound), axis=1)]
        self.natoms = len(self.positions)

        self.remained_per_step = []
        self.released_per_step = []

    def get_gamma(self, energy: float = None, temperature: float = None):
        r = 8.31446261815324  # J / (K * mol)
        T = self.current_temp if temperature is None else temperature
        e = self.energy if energy is None else energy
        alpha = 1 - self.density
        p1 = math.exp(- e / (r * T))
        p2 = math.exp(- e / 2 / (r * T))  # p2是给定活化能的一半，用于假定晶面扩散的活化能
        gamma = p1 * alpha * (1 - self.fracture) + self.fracture * p2 + self.boundary + self.defect
        return gamma

    def get_natoms(self, pos=None):
        if pos is None:
            return len(self.positions)
        pos = pos[np.all((pos >= self.min_bound) & (pos <= self.max_bound), axis=1)]
        for dom in self.inclusions:
            pos = pos[~np.all((pos >= dom.min_bound) & (pos <= dom.max_bound), axis=1)]
        return len(pos)

    def set_attr(self, name, value, index=None):
        if index is None:
            self.attr.update({name: value})
        else:
            if name not in self.attr.keys():
                self.attr.update({name: {}})
            self.attr[name].update({index: value})


class DiffSimulation:
    def __init__(self):
        self.name = "example"
        self.loc = "examples"
        self.score = 0
        self.end_at = 0
        self.dimension = 3

        self.size_scale = 1
        self.length_scale = 1

        self.atom_density = 1e14  # 原子密度 个/立方厘米
        self.frequency = 1e13  # 每秒振动的次数, Debye frequency

        self.grain_size = 300  # 0.3 mm = 300 μm = 300000 nm
        self.grid_size = self.grain_size * self.size_scale  # 网格大小
        self.nsteps: int = ...
        self.natoms: int = ...
        self.epsilon = 1e-5
        self.step_length = 0.0001 * 2.5  # 1 A = 0.1 nm = 0.0001 μm

        self.seq: [List, np.ndarray] = []
        self.domains: List[Domain] = []
        self.positions: [List, np.ndarray] = []

        self.remained_per_step = []
        self.released_per_step = []

    def setup(self):
        """
        Used to set or reset positions of atoms
        :return:
        """
        # 网格大小和颗粒边界
        self.grid_size = self.grain_size * self.size_scale  # 网格大小
        center = np.array([0 for i in range(self.dimension)])
        self.min_bound = center - self.grid_size / 2
        self.max_bound = center + self.grid_size / 2

        # 粒子位置
        self.positions = np.concatenate([dom.positions for dom in self.domains])
        # 粒子数目
        self.natoms = len(self.positions)

        self.remained_per_step = []
        self.released_per_step = []

    def run_sequence(self, times, temperatures, statuses, targets, domains: List[Domain], nsteps_factor=None,
                     simulating: bool = False, epsilon=0.001, start_time=0, k=1.2, use_walker1=True,
                     scoring: bool = False, decay=0, parent=0):
        pos = copy.deepcopy(self.positions)
        self.seq = []  # 用于记录参数

        for index, (duration, temperature, target, status) in enumerate(zip(times, temperatures, targets, statuses)):

            self.seq.append({})

            loop = 0
            e_low = 100 * 1000
            e_up = 300 * 1000

            _start = time.time()
            heating_duration = duration - start_time  # in seconds
            start_time = duration
            temperature = temperature + 273.15  # in Kelvin

            while loop < 50:

                e = (e_low + e_up) / 2

                total_steps = self.frequency * heating_duration  # 振动次数
                conditions = np.array(
                    [[*dom.min_bound, *dom.max_bound,
                      dom.get_gamma(temperature=temperature, energy=e if simulating else None)] for dom in domains if dom.energy != 0 and not simulating])

                if use_walker1:

                    # print(f"调整前: {nsteps_factor = }, gamma = {conditions[0][-1]}, {total_steps = }")
                    nsteps_factor = 10 ** math.floor(
                        -math.log10(max(conditions[:, -1]))) * 1000  # 用来调节次数和步长的巨大差异，次数太大，但步程太小
                    while int(total_steps / nsteps_factor) < 1:
                        nsteps_factor /= 10
                    while int(total_steps / nsteps_factor) > 10000:
                        nsteps_factor *= 10
                    total_steps = int(total_steps / nsteps_factor)  # 振动次数
                    conditions[:, -1] *= nsteps_factor
                    compensation = max(3, *(np.ceil(np.sqrt(conditions[:, -1])) * 3))
                    boundary_factor = 1
                    if max(conditions[:, -1]) > 1000:
                        # k = 0时，bf = 1, 完全无缩放 # k = 1时，bf <= 0.5 小步行走将至少缩放一半   # 推荐值 k = 1.2
                        boundary_factor = 0.1 ** (k * math.log10(1 + (max(conditions[:, -1]) // 1000)))
                    step_length = self.step_length / np.sqrt(pos.shape[1] if len(pos.shape) > 1 else 1)
                    scale = int(total_steps)
                    # print(f"调整后: {nsteps_factor = }, gamma = {conditions[0][-1]}, {total_steps = }, {compensation = }, {boundary_factor = }")

                    _pos = walker(
                        copy.deepcopy(pos), step_length=step_length, total_nsteps=total_steps,
                        min_bound=self.min_bound, max_bound=self.max_bound, scale=scale,
                        compensation=compensation, remove=True, conditions=conditions, boundary_factor=boundary_factor,
                        decay=decay, parent=parent
                    )

                    # self.seq[index].update({
                    #     "duration": duration, "temperature": temperature, "target": target, "step_length": step_length,
                    #     "scale": scale, "compensation": compensation, "boundary_factor": boundary_factor, "e": e,
                    #     "simulation": simulating, "nsteps_factor": nsteps_factor, "conditions": conditions
                    # })

                else:
                    step_length = self.step_length / np.sqrt(pos.shape[1] if len(pos.shape) > 1 else 1)
                    scale = k
                    _pos = walker2(copy.deepcopy(pos), duration=heating_duration, step_length=step_length,
                                   min_bound=self.min_bound, max_bound=self.max_bound, time_scale=scale,
                                   frequency=self.frequency, conditions=conditions, remove=True, decay=decay, parent=parent)
                    boundary_factor = 1
                    compensation = 0

                self.seq[index].update({
                    "duration": duration, "temperature": temperature, "target": target, "step_length": step_length,
                    "scale": scale, "compensation": compensation, "boundary_factor": boundary_factor, "e": e,
                    "simulation": simulating, "nsteps_factor": nsteps_factor, "conditions": conditions
                })

                released = 1 - len(_pos) / self.natoms
                d = released - target
                if simulating:
                    loop += 1
                    print(f"{e = }, {d = }, {abs(d) <= epsilon}")
                    if abs(d) <= epsilon:
                        break
                    elif d > 0:
                        e_low = e
                    else:
                        e_up = e
                elif scoring:
                    self.score += (d * 100) ** 2
                    print(f"{self.score = }, {d = }")
                    if self.score > 4 * (index + 1):
                        raise OverEpsilonError(index, self.score, 4 * (index + 1))
                    else:
                        break
                else:
                    break

            pos = _pos

            if status:
                # status == True: argon collection, False: pumping
                for dom in self.domains:
                    _c = dom.get_natoms(pos)
                    dom.remained_per_step.append(_c)
                    dom.released_per_step.append(dom.natoms - _c)
                    dom.set_attr("energy", float(e), index=index)

                self.remained_per_step.append(len(pos))
                self.released_per_step.append(self.natoms - len(pos))

            print(f"{index = } {duration} - {heating_duration = } - {temperature = } - {total_steps = } - conc = {len(pos) / self.natoms * 100:.2f}% - {time.time() - _start:.5f}s")

        self.positions = copy.deepcopy(pos)

    def run_persecond(self, times, temperatures, statuses, targets, domains: List[Domain], scale=None,
                     simulation: bool = False, epsilon=0.001, start_time=0):
        seq = []
        for i in range(1, int(times[-1] + 1), 300):
            for index, time in enumerate(times):
                if (times[index-1] if index > 0 else 0) < i <= times[index]:
                    seq.append([i, temperatures[index], statuses[index], targets[index]])
        seq = np.transpose(seq)
        return self.run_sequence(*seq, domains=domains, nsteps_factor=scale, simulating=simulation,
                                 epsilon=epsilon, start_time=start_time)


def run(times, temps, statuses, energies, fractions, ndoms: int = 1, grain_szie=275, atom_density=1e10, frequency=1e13,
        dimension: int = 3, targets: list = None, epsilon: float = 0.001, simulation: bool = False,
        file_name: str = "Y70", ignore_error: bool = False, positions=None, **kwargs):
    """
    :param times:
    :param temps:
    :param statuses:
    :param energies:
    :param fractions:
    :param dimension:
    :param ndoms:
    :param targets:
    :param epsilon:
    :param simulation:
    :param file_name:
    :param ignore_error:
    :return:
    """

    demo = DiffSimulation()

    def _(n, es, fs):
        # fs 应从大到小，父空间在前，子空间在后

        # demo.grain_size = 300
        # demo.size_scale = 0.05
        # demo.atom_density = 1e14  # 原子密度 个/立方厘米
        demo.dimension = dimension

        demo.size_scale = 1
        demo.grain_size = grain_szie
        demo.atom_density = atom_density  # 原子密度 个/立方厘米
        demo.frequency = frequency

        # domains应该从外到内
        domains = []
        for i in range(n-1, 0-1, -1):
            size = int(demo.grain_size * fs[i]) * demo.size_scale
            center = np.zeros(demo.dimension)
            if i == 2:
                ad = demo.atom_density * 5 / 4
            else:
                ad = demo.atom_density
            dom = Domain(
                dimension=demo.dimension, atom_density=ad, min_bound=center - size / 2, max_bound=center + size / 2,
                energy=es[i], fraction=fs[i], inclusions=[domains[-1]] if len(domains) >= 1 else []
            )
            domains.append(dom)
        # domains应该从外到内, 上面为了inclusion以及方便不同扩散域设置不同的密度，要按照从小到大的顺序生成，但是后面行走的时候要根据不同条件设置系数，要从外到内
        demo.domains = sorted(domains, key=lambda dom: dom.fraction, reverse=True)

        demo.setup()

        demo.name = f"{file_name}"

        print(f"Total Atoms: {demo.natoms}, atoms in each dom: {[dom.natoms for dom in demo.domains]} filename: {demo.name}")

        if positions is not None:
            demo.positions = positions
            demo.natoms = len(positions)

        demo.run_sequence(times=times, temperatures=temps, statuses=statuses, targets=targets, domains=demo.domains,
                          epsilon=epsilon, simulating=simulation, **kwargs)
        # demo.run_persecond(times=times, temperatures=temps, domains=demo.domains, targets=target,
        #                    epsilon=epsilon, simulation=simulation)

        return demo

    try:
        return _(ndoms, energies, fractions), True
    except OverEpsilonError as e:
        if ignore_error:
            return demo, False
        else:
            raise


def run_40ar(times, temps, statuses, energies, fractions, ndoms: int = 1, grain_szie=275, atom_density=1e10, frequency=1e13,
        dimension: int = 3, targets: list = None, epsilon: float = 0.001, simulation: bool = False,
        file_name: str = "Y70", ignore_error: bool = False, **kwargs):
    """
    :param times:
    :param temps:
    :param statuses:
    :param energies:
    :param fractions:
    :param dimension:
    :param ndoms:
    :param targets:
    :param epsilon:
    :param simulation:
    :param file_name:
    :param ignore_error:
    :return:
    """

    demo = DiffSimulation()

    def _(n, es, fs):
        # fs 应从大到小，父空间在前，子空间在后

        # demo.grain_size = 300
        # demo.size_scale = 0.05
        # demo.atom_density = 1e14  # 原子密度 个/立方厘米
        demo.dimension = dimension

        demo.size_scale = 1
        demo.grain_size = grain_szie
        demo.atom_density = atom_density  # 原子密度 个/立方厘米
        demo.frequency = frequency

        # domains应该从外到内
        domains = []
        for i in range(n-1, 0-1, -1):
            size = int(demo.grain_size * fs[i]) * demo.size_scale
            center = np.zeros(demo.dimension)
            dom = Domain(
                dimension=demo.dimension, atom_density=demo.atom_density, min_bound=center - size / 2, max_bound=center + size / 2,
                energy=es[i], fraction=fs[i], inclusions=[domains[-1]] if len(domains) >= 1 else []
            )
            domains.append(dom)
        # domains应该从外到内, 上面为了inclusion以及方便不同扩散域设置不同的密度，要按照从小到大的顺序生成，但是后面行走的时候要根据不同条件设置系数，要从外到内
        demo.domains = sorted(domains, key=lambda dom: dom.fraction, reverse=True)

        demo.setup()

        demo.name = f"{file_name}"

        print(f"Total Atoms: {demo.natoms}, atoms in each dom: {[dom.natoms for dom in demo.domains]} filename: {demo.name}")

        demo.run_sequence(times=times, temperatures=temps, statuses=statuses, targets=targets, domains=demo.domains,
                          epsilon=epsilon, simulating=simulation, **kwargs)
        # demo.run_persecond(times=times, temperatures=temps, domains=demo.domains, targets=target,
        #                    epsilon=epsilon, simulation=simulation)

        return demo

    try:
        return _(ndoms, energies, fractions), True
    except OverEpsilonError as e:
        if ignore_error:
            return demo, False
        else:
            raise


def save_ads(ads: DiffSimulation, dir_path: str = None, name: str = None):
    if dir_path is None:
        dir_path = ""
    if name is None:
        name = ads.name
    if not name.endswith(".ads"):
        name = f"{name}.ads"
    dir_path = os.path.join(dir_path, name)
    with open(dir_path, 'wb') as f:
        f.write(pickle.dumps(ads))
    return name


def read_ads(file_path: str) -> DiffSimulation:

    class RenameUnpickler(pickle.Unpickler):
        def find_class(self, module: str, name: str):
            if "argon_diffusion_simulator" in module:
                return super(RenameUnpickler, self).find_class(DiffSimulation().__module__, name)
            else:
                return super(RenameUnpickler, self).find_class(module, name)

    def renamed_load(file_obj) -> DiffSimulation:
        return RenameUnpickler(file_obj).load()

    try:
        with open(file_path, 'rb') as f:
            return renamed_load(f)
    except (Exception, BaseException):
        raise ValueError(f"ModuleNotFoundError")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - calc - histogram
# ==========================================
#
#
"""


import numpy as np


def get_data(x: list, s: float = None, r: str = 'sturges', w: float = None, c: int = None):
    """

    Parameters
    ----------
    x : input data to yield bins
    s : starting point
    r : rules, string
    w : bin width or interval, float for specific number
    c : bin count

    Returns
    -------
    tuple. counts: [number of points in each bin], bins: [half bins], s, r, w, c, e, res: [values in each bin]
    """
    if len(x) == 0 or max(x) == min(x):
        return None
    else:
        x = [round(xi, 2) for xi in x]
    if isinstance(r, str) and r.lower().find('square-root') != -1:
        # Square-root choice, used by Excel's Analysis Toolpak histograms and many other
        c = np.ceil(len(x) ** (1. / 2.))
    if isinstance(r, str) and r.lower().find('sturges') != -1:
        # Sturges' formula, Ceiling(log2n)
        c = np.ceil(np.log2(len(x))) + 1
    if isinstance(r, str) and r.lower().find('rice') != -1:
        # Rice Rule
        c = np.ceil(2 * len(x) ** (1. / 3.))
    if isinstance(r, str) and r.lower().find('scott') != -1:
        # Scott's normal reference rule, optimal for random samples of normally distributed data
        w = np.ceil(3.49 * (sum([float(i - sum(x) / len(x)) ** 2 for i in x]) / len(x)) ** (1. / 2.) / len(x) ** (1. / 3.))
        d = 0.5 * 10 ** (int(np.log(w)) - 1)
    if w is None and c is None:
        return get_data(x=x, s=s, w=w, c=c, r='sturges')

    if s is None and isinstance(c, (int, np.float64)):
        d = 0.5 * 10 ** int(np.log(abs(max(x) - min(x)) / c))
    if s is None:
        d = d if min(x) - d >= 0 else min(x)
        s = round(min(x) - d, 2)
        e = round(max(x) + d, 2)
    else:
        e = round(max(x) + min(x) - s, 2)
    if isinstance(c, (int, np.float64)) and not isinstance(w, (int, float, np.float64)):
        w = round(abs(e - s) / c, 2)
    bins = [s + i * w for i in range(10000) if s + (i - 1) * w <= max(x)]
    bins = list(set(bins))
    bins.sort()
    c = len(bins) - 1
    counts = [0] * c
    res = [[]] * c
    half_bins = [round((bins[i] + bins[i + 1]) / 2, 2) for i in range(c)]
    bin_ranges = [[]] * c
    for i in range(c):
        bin_ranges[i] = [round(bins[i], 2), round(bins[i + 1], 2)]
        for xi in x:
            if bins[i] <= xi < bins[i + 1] or bins[i] <= xi <= bins[i + 1] and i == c - 1:
                counts[i] += 1
                res[i].append(xi)

    return counts, half_bins, bin_ranges, s, r, w, c, bins[-1], res


def get_kde(x: list, h: (float, int) = None, a: str = None, k: str = 'normal',
            s: (float, int) = None, e: (float, int) = None, n: int = 200):
    """

    Parameters
    ----------
    x : input data
    h : bandwidth
    a : auto width rule
    k : kernel function name, default is normal, standard normal density function
    s : KDE curve starting point
    e : KDE curve ending point
    n : points number of KDE line

    Returns
    -------

    """
    x = [round(xi, 2) for xi in x]
    x.sort()

    def get_uniform_x(_x: list, _s=s, _e=e, _np=n):
        _s = min(_x) if _s is None else _s
        _e = max(_x) if _e is None else _e
        _line_x = []
        _step = abs(_e - _s) / _np
        _line_x = [_s + i * _step for i in range(_np)] + [_e]
        _line_x.sort()
        return _line_x

    def h_scott(_x, _se):  # Scott, 1992
        return 1.06 * _se * len(_x) ** (-1. / 5.)

    def h_silverman(_x, _se):  # Silverman, 1986
        return 0.9 * min(_se, (_x[int(3 / 4 * len(_x))] - _x[int(1 / 4 * len(_x))]) / 1.34) * len(_x) ** (-1. / 5.)

    # Normal function
    def k_normal(_xi, _u=0, _se=1, _h=h):
        return 1 / (_se * np.sqrt(2 * np.pi)) * (np.exp(-1. / 2. * ((_xi - _u) / (_h * _se)) ** 2))

    def k_epanechnikov(_xi, _u=0, _h=h):
        _xi = (_xi - _u) / _h
        return 3 / 4 * (1 - _xi ** 2) if abs(_xi) <= 1 else 0

    def k_uniform(_xi, _u=0, _h=h):
        _xi = (_xi - _u) / _h
        return 1 / 2 if abs(_xi) <= 1 else 0

    def k_triangular(_xi, _u=0, _h=h):
        _xi = (_xi - _u) / _h
        return 1 - abs(_xi) if abs(_xi) <= 1 else 0

    mean_x = sum(x) / len(x)
    se = np.sqrt(sum([(xi - mean_x) ** 2 for xi in x]) / (len(x) - 1))

    if (a is None or a == 'none') and (h is None or h <= 0):
        # Default rule for h is Scott's rule
        a = 'Scott'
    if isinstance(a, str):
        if a.lower() == 'scott':
            h = h_scott(x, se)
        elif a.lower() == 'silverman':
            h = h_silverman(x, se)
        else:
            a = 'none'
    else:
        a = 'none'

    # Get points that are evenly distributed over the range (min_x, max_x) to get a KDE curve
    line_x = get_uniform_x(_x=x, _s=s, _e=e)

    if k.lower() == 'normal':
        k_normal_res = [[k_normal(_xi, _u=xi, _h=h) for _xi in line_x] for xi in x]
    elif k.lower() == 'epanechnikov':
        k_normal_res = [[k_epanechnikov(_xi, _u=xi, _h=h) for _xi in line_x] for xi in x]
    elif k.lower() == 'uniform':
        k_normal_res = [[k_uniform(_xi, _u=xi, _h=h) for _xi in line_x] for xi in x]
    elif k.lower() == 'triangular':
        k_normal_res = [[k_triangular(_xi, _u=xi, _h=h) for _xi in line_x] for xi in x]
    else:
        return get_kde(x=x, h=h, a=a, k='normal', s=s, e=e, n=n)

    res = [
        sum([k_normal_res[j][i] for j in range(len(x))]) / (len(x) * h) for i in range(len(line_x))
    ]

    return [[line_x, res], h, k, a]




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# ==========================================
# Copyright 2023 Yang 
# ararpy - setup
# ==========================================
#
#
#

import setuptools
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / 'README.md').read_text(encoding='utf-8')

setuptools.setup(
    name='ararpy',  #
    version='0.1.195',  # version
    author='Yang Wu',
    author_email='wuycug@hotmail.com',
    description='A project for Ar-Ar geochronology',  # short description
    long_description=long_description,  # detailed description in README.md
    long_description_content_type='text/markdown',
    url='https://github.com/wuyangchn/ararpy.git',  # github url
    packages=setuptools.find_packages(),
    package_data={'ararpy': ['examples/*']},
    # install_requires=[
    #     "chardet", "numpy", "OriginExt", "pandas", "parse", "python_dateutil",
    #     "scipy", "setuptools", "xlrd", "XlsxWriter", "pdf_maker"
    # ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    license='MIT',
    python_requires='>=3.5',
)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - smp - basic
# ==========================================
#
#
#
"""
# === Internal imports ===
import os
import re
import traceback
import pandas as pd
import numpy as np
import copy
from typing import Optional, Union, List
from .. import calc
from ..files.basic import (read as read_params)
from .sample import Sample, Table, Plot, ArArData, ArArBasic, Sequence, RawData
from .initial import preference_keys

Set = Plot.Set
Label = Plot.Label
Axis = Plot.Axis
Text = Plot.Text

pd.options.mode.chained_assignment = None  # default='warn'


# =======================
# Calculate ages
# =======================
def calc_apparent_ages(smp: Sample):
    """

    Parameters
    ----------
    smp

    Returns
    -------

    """
    if smp.Info.sample.type == "Unknown":
        age = calc_age(smp=smp)
        smp.ApparentAgeValues[2:6] = age
        smp.PublishValues[5:7] = copy.deepcopy(age[0:2])
    if smp.Info.sample.type == "Standard":
        j = calc_j(smp=smp)
        smp.ApparentAgeValues[2:4] = j
        smp.PublishValues[5:7] = copy.deepcopy(j[0:2])
    if smp.Info.sample.type == "Air":
        mdf = calc_mdf(smp=smp)
        smp.ApparentAgeValues[2:4] = mdf
        smp.PublishValues[5:7] = copy.deepcopy(mdf[0:2])


def calc_j(ar40ar39=None, params: dict = None, smp: Sample = None, index: list = None):
    std_age, std_err = smp.TotalParam[59:61]
    r, sr = smp.ApparentAgeValues[0:2]  # ar40ar39, error
    f, rsf = smp.TotalParam[34:36]  # L, sL
    j, sj = calc.jvalue.j_value(std_age, std_err, r, sr, f, rsf)
    return [j.tolist(), sj.tolist()]


def calc_mdf(ar40ar36=None, params: dict = None, smp: Sample = None, index: list = None):
    std_air, std_err = smp.TotalParam[93:95]
    m36, sm36 = smp.TotalParam[71:73]
    m40, sm40 = smp.TotalParam[79:81]
    air, sair = smp.ApparentAgeValues[0:2]  # ar40ar36 air, error
    discrimination_method = smp.TotalParam[100]  # L, sL
    mdf = []
    for i in range(len(std_air)):
        k = calc.corr.mdf(air[i], sair[i], m36[i], m40[i], std_air[i], std_err[i])  # linear, exp, pow
        mdf.append({"linear": k[0:2], "exp": k[2:4], "pow": k[4:6]}[discrimination_method[i].lower()])
    return np.transpose(mdf).tolist()


def calc_age(ar40ar39=None, params: dict = None, smp: Sample = None, index: list = None):
    """
    40Ar/39Ar age calculation, two methods are supported: Min et al. (2000) or general equation.
    Parameters
    ----------
    ar40ar39 : 2D DataFrame, Series, list
    params : dict
    smp : Sample instance
    index:

    Returns
    -------

    """
    params_index_dict = {
        34: 'L', 35: 'sL', 36: 'Le', 37: 'sLe', 38: 'Lb', 39: 'sLb', 48: 'A', 49: 'sA',
        50: 'Ae', 51: 'sAe', 52: 'Ab', 53: 'sAb', 59: 't', 60: 'st', 67: 'J', 68: 'sJ',
        81: 'W', 82: 'sW', 83: 'No', 84: 'sNo', 85: 'Y', 86: 'sY', 87: 'f', 88: 'sf', 110: 'Min'
    }

    if isinstance(ar40ar39, pd.Series):
        ar40ar39 = [ar40ar39.tolist(), [0] * ar40ar39.size]
    if isinstance(ar40ar39, pd.DataFrame):
        ar40ar39 = ar40ar39.transpose().values.tolist()
    if ar40ar39 is None and smp is not None:
        ar40ar39 = smp.ApparentAgeValues[0:2]
    if len(np.shape(ar40ar39)) == 1:
        ar40ar39 = np.reshape(ar40ar39, (2, 1))
    if index is None:
        index = list(range(np.shape(ar40ar39)[-1]))

    if smp is None and params is None:
        raise ValueError(f"Parameters are required for calculating ages, or it is empty.")
    if params is not None:
        for key, val in params.items():
            if not isinstance(val, list):
                params[key] = list(val)
    if smp is not None:
        try:
            params_from_smp = dict(zip(
                list(params_index_dict.values()),
                [[smp.TotalParam[i][j] for j in index] for i in params_index_dict.keys()]
            ))
        except Exception:
            print(traceback.format_exc())
            raise ValueError(f"Parameters cannot be found in the given sample object")
        if params is not None:
            params_from_smp.update(params)
        params = params_from_smp

    u = 'Ma'
    try:
        u = smp.Info.preference['ageUnit']
    except:
        print(traceback.format_exc())
        pass
    finally:
        if u.lower() == 'ga':
            age_unit_factor = 1000000000
        elif u.lower() == 'ma':
            age_unit_factor = 1000000
        elif u.lower() == 'ka':
            age_unit_factor = 1000
        else:
            age_unit_factor = 1

    # check if using Min equation
    params['Min'] = [i if isinstance(i, bool) else False for i in params['Min']]

    idx1 = np.flatnonzero(np.where(params['Min'], True, False))  # True, using Min equation
    idx2 = np.flatnonzero(np.where(params['Min'], False, True))  # False
    k1, k2 = [], []
    if np.size(idx1) > 0:
        k1 = calc.age.calc_age_min(
            F=[ar40ar39[0][i] for i in idx1], sF=[ar40ar39[1][i] for i in idx1],
            **dict(zip(params.keys(), [[val[i] for i in idx1] for val in params.values()])))
        k1 = [i / age_unit_factor for i in k1]
    if np.size(idx2) > 0:
        k2 = calc.age.calc_age_general(
            F=[ar40ar39[0][i] for i in idx2], sF=[ar40ar39[1][i] for i in idx2],
            **dict(zip(params.keys(), [[val[i] for i in idx2] for val in params.values()])))
        k2 = [i / age_unit_factor for i in k2]

    # idx1 = params[params['Min'].astype(bool)].index
    # idx2 = params[~params['Min'].astype(bool)].index  # The operators are: | for or, & for and, and ~ for not
    # k1, k2 = [], []
    # if idx1.size > 0:
    #     k1 = calc.age.calc_age_min(
    #         F=ar40ar39.take(idx1)[0], sF=ar40ar39.take(idx1)[1], **params.take(idx1).to_dict('list'))
    # if idx2.size > 0:
    #     k2 = calc.age.calc_age_general(
    #         F=ar40ar39.take(idx2)[0], sF=ar40ar39.take(idx2)[1], **params.take(idx2).to_dict('list'))

    columns = ['age', 's1', 's2', 's3']
    ages = pd.concat([
        pd.DataFrame([*k1], columns=idx1, index=columns),
        pd.DataFrame([*k2], columns=idx2, index=columns)
    ], axis=1).transpose().reset_index(drop=True)

    if len(index) == 1:
        return ages.transpose().squeeze().tolist()
    else:
        return ages.transpose().values.tolist()


# =======================
# Search components
# =======================
def get_content_dict(smp: Sample):
    """

    Parameters
    ----------
    smp

    Returns
    -------

    """
    return dict(zip(
        ['smp', 'blk', 'cor', 'deg', 'pub', 'age', 'iso', 'pam', 'mak', 'seq'],
        [smp.SampleIntercept, smp.BlankIntercept, smp.CorrectedValues, smp.DegasValues,
         smp.PublishValues, smp.ApparentAgeValues, smp.IsochronValues, smp.TotalParam,
         [smp.IsochronMark], [smp.SequenceName, smp.SequenceValue]]
    ))


def get_dict_from_obj(obj: (Sample, Plot, Table, Set, Label, Axis, Text,
                            ArArBasic, ArArData, Sequence, RawData)):
    """

    Parameters
    ----------
    obj

    Returns
    -------

    """
    res = {}
    for key, attr in obj.__dict__.items():
        if not isinstance(attr, (Sample, Plot, Table, Set, Label, Axis, Text,
                                 ArArBasic, ArArData, Sequence, RawData)):
            res.update({key: attr})
        else:
            res.update({key: get_dict_from_obj(attr)})
    return res


def get_components(smp: Sample):
    """
    Get updated sample.Components dict
    Parameters
    ----------
    smp

    Returns
    -------

    """
    components_name = [
        '0', '1', '2', '3', '4', '5', '6', '7', '8',
        'figure_1', 'figure_2', 'figure_3', 'figure_4', 'figure_5', 'figure_6', 'figure_7', 'figure_8', 'figure_9',
    ]
    components = {}
    for key in components_name:
        comp = get_component_byid(smp, key)
        components.update({key: comp})
    return components


def get_component_byid(smp: Sample, comp_id: str):
    """
    Get a component (Table or Plot) based on input id
    Parameters
    ----------
    smp
    comp_id

    Returns
    -------

    """
    for key, val in smp.__dict__.items():
        if isinstance(val, (Plot, Table, ArArBasic)) and getattr(val, 'id') == comp_id:
            return val


def get_plot_set(plot: Plot, comp_id):
    """
    Get a Set, Text, Axis, Label of a sample instance based on given id
    """
    for v in [getattr(plot, k) for k in dir(plot)]:
        if isinstance(v, Plot.BasicAttr) and v.id.lower() == comp_id.lower():
            return v
    return None


# =======================
# Update
# =======================
def update_plot_from_dict(plot, attrs: dict):
    """ Set instance based on the given attrs dictionary
    Parameters
    ----------
    plot
    attrs : dict
        for example: update_plot_from_dict(sample.Info, info: dict), where info = {
            'sample': {'name': 'name', 'material': 'material', 'location': 'location'},
            'researcher': {'name': 'researcher'},
            'laboratory': {'name': 'laboratory'}
        }

    Returns
    -------

    """

    def _do(_plot, _attrs: dict):
        for k1, v1 in _attrs.items():
            if isinstance(v1, dict):
                if hasattr(_plot, k1):
                    if isinstance(getattr(_plot, k1), dict):
                        setattr(_plot, k1, calc.basic.update_dicts(getattr(_plot, k1), v1))
                        # setattr(_plot, k1, v1)
                    else:
                        _do(getattr(_plot, k1), v1)
            else:
                setattr(_plot, k1, v1)

    _do(_plot=plot, _attrs=attrs)
    return plot


def update_object_from_dict(obj, attrs: dict):
    """
    update object
    Parameters
    ----------
    obj
    attrs

    Returns
    -------

    """
    def _do(_obj, _attrs: dict):

        for k1, v1 in _attrs.items():
            if hasattr(_obj, k1):
                if getattr(_obj, k1) == v1:
                    continue
                elif isinstance(v1, dict):
                    if isinstance(getattr(_obj, k1), dict):
                        setattr(_obj, k1, calc.basic.update_dicts(getattr(_obj, k1), v1))
                        # setattr(_plot, k1, v1)
                    else:
                        _do(getattr(_obj, k1), v1)
                else:
                    setattr(_obj, k1, v1)
            else:
                setattr(_obj, k1, v1)

    _do(_obj=obj, _attrs=attrs)
    return obj


# =======================
# Merge sample instances
# =======================
def get_merged_smp(a: Sample, b: (Sample, dict)):
    """ Comparing two sample instances a and b
        This function is used to update sample instance to make sure JS can read properties it required.
    Parameters
    ----------
    a : sample instance that has old attributes,
    b : new sample instance that has some new attributes or a has similar but different name for this attribute

    Returns
    -------
    None
        return none, but a will be updated after calling this function

            for example:
                A = Sample(id = 'a', set1 = Set(id = 'set1', data = [], symbolSize = 10)),
                B = Sample(id = 'b', set1 = Set(id = 'set1', data = [2023], symbol_size = 5, line_type = 'solid')),
                after get_merged_smp(A, B), A will be Sample(id = 'a', 'set1' = Set(id = 'set1', data = [],
                                                        symbol_size = 10, line_type = 'solid'))
    """

    def get_similar_name(_name: str):
        res = []
        for i in range(len(_name) + 1):
            str_list = [i for i in _name]
            str_list.insert(i, '_')
            res.append(''.join(str_list))
        for i in range(len(_name)):
            str_list = [i for i in _name]
            str_list[i] = str_list[i].capitalize()
            res.append(''.join(str_list))
        for i in range(len(_name)):
            str_list = [i for i in _name]
            if _name[i] in '-_':
                str_list.pop(i)
                res.append(''.join(str_list))
        return res

    if not isinstance(b, dict):
        b = b.__dict__

    for name, attr in b.items():
        if hasattr(a, name):
            if isinstance(attr, (Plot, Table, ArArBasic, Plot.BasicAttr)):
                if not type(getattr(a, name)) == type(attr):
                    if isinstance(getattr(a, name), dict):
                        setattr(a, name, type(attr)(**getattr(a, name)))
                    else:
                        setattr(a, name, type(attr)())
                get_merged_smp(getattr(a, name), attr)
            if isinstance(attr, dict) and isinstance(getattr(a, name), dict):
                setattr(a, name, calc.basic.merge_dicts(getattr(a, name), attr))
            if isinstance(attr, list) and isinstance(getattr(a, name), list):
                if len(attr) > len(getattr(a, name)):
                    setattr(a, name, getattr(a, name) + attr[len(getattr(a, name)):])
            continue
        else:
            for xxx in get_similar_name(name):
                for xx in get_similar_name(xxx):
                    for x in get_similar_name(xx):
                        if hasattr(a, x):
                            # print(f'Has similar {name} = {x}: {getattr(a, x)}')
                            setattr(a, name, getattr(a, x))
                            break
                    else:
                        continue
                    break
                else:
                    continue
                break
            if not hasattr(a, name):
                setattr(a, name, attr)


# =======================
# Difference between two sample instances
# =======================
def get_diff_smp(backup: (dict, Sample), smp: (dict, Sample)):
    """ Comparing two sample component dicts or sample instances, and return difference between them.
    Parameters
    ----------
    backup : backup of sample.Components or sample before changed.
    smp : sample.Components or sample after changed

    Returns
    -------
    dict
        dict of keys and values that have difference, iterate to a sepcial difference,

            for example:
                A = Sample(id = 'a', set1 = Set(id = 'set1', data = [])),
                B = Sample(id = 'b', set1 = Set(id = 'set1', data = [2023])),
                res = get_diff_smp(A, B) will be {'id': 'b', 'set1': {'data': [2023]}}

    """
    res = {}
    if isinstance(backup, Sample) and isinstance(smp, Sample):
        return get_diff_smp(backup.__dict__, smp.__dict__)
    for name, attr in smp.items():
        if name not in backup.keys() or not isinstance(backup[name], type(attr)):
            res.update({name: attr})
            continue
        if isinstance(attr, dict):
            # if name not in backup.keys():
            #     res.update({name: attr})
            #     continue
            _res = get_diff_smp(backup[name], attr)
            if _res != {}:
                res.update({name: _res})
            continue
        if isinstance(attr, np.ndarray):
            if not np.array_equal(attr, backup[name]):
                res.update({name: attr})
            continue
        if isinstance(attr, (Plot, Table, Plot.Text, Plot.Axis, Plot.Set,
                             Plot.Label, Plot.BasicAttr, ArArBasic, ArArData)):
            _res = get_diff_smp(backup[name].__dict__, attr.__dict__)
            if _res != {}:
                res.update({name: _res})
            continue
        if str(backup[name]) == str(attr) or backup[name] == attr:
            continue
        res.update({name: attr})
    return res


# =======================
# Set parameters
# =======================
def set_params(smp: Sample, params: Union[List, str], flag: Optional[str] = None, rows: Optional[List] =None):
    """
    Parameters
    ----------
    smp
    params
    flag : optional, should be one of 'calc', 'irra', and 'smp'. If it is not given,
        the text of the extension without a dot will be used
    rows

    Returns
    -------

    """
    if rows is None:
        rows = []
    rows = list(set(rows))

    if isinstance(params, str) and os.path.isfile(params):
        if flag is None:
            flag = params.split(".")[-1]
        return set_params(smp, read_params(params), flag=flag)

    def isValid(item):
        valid = True
        if isinstance(item, (float, int)):
            valid = not np.isnan(item)
        return valid and item is not None and item != ""

    def remove_none(old_params, new_params, row_list, length):
        res = [[old_params[i][j] if j < len(old_params[i]) else None for j in range(max(len(old_params[i]), length))] for i in range(len(old_params))]
        for index, item in enumerate(new_params):
            if isValid(item):
                for row in row_list:
                    res[index][row] = item
        return res

    n = len(smp.SequenceName)
    if n == 0:
        raise ValueError(f"The number of sample sequences is zero")

    if flag == 'calc':
        smp.TotalParam[34:56] = remove_none(smp.TotalParam[34:56], params[0:22], rows, n)
        smp.TotalParam[71:97] = remove_none(smp.TotalParam[71:97], params[22:48], rows, n)
    elif flag == 'irra':
        smp.TotalParam[0:20] = remove_none(smp.TotalParam[0:20], params[0:20], rows, n)
        smp.TotalParam[56:58] = remove_none(smp.TotalParam[56:58], params[20:22], rows, n)  # Cl36/38 productivity
        smp.TotalParam[20:27] = remove_none(smp.TotalParam[20:27], params[22:29], rows, n)
        irradiation_time = []
        duration = []
        if None not in params[29:-3] and '' not in params[29:-3]:
            for i in range(len(params[29:-3])):
                if i % 2 == 0:
                    irradiation_time.append(params[29:-3][i] + 'D' + str(params[29:-3][i + 1]))
                    duration.append(float(params[29:-3][i + 1]))
            smp.TotalParam[27:30] = remove_none(smp.TotalParam[27:30], ['S'.join(irradiation_time), params[-3], sum(duration)], rows, n)
        if params[-5] != '':
            smp.TotalParam[30:32] = remove_none(smp.TotalParam[30:32], [params[-5], None], rows, n)
        try:
            stand_time_second = [
                calc.basic.get_datetime(*re.findall(r"\d+", smp.TotalParam[31][i])) - calc.basic.get_datetime(
                    *re.findall(r"\d+", smp.TotalParam[30][i])) for i in range(len(smp.TotalParam[30]))]
        except (BaseException, Exception):
            print(f'Error in calculate standing duration: {traceback.format_exc()}')
            smp.TotalParam[32] = [None for index in range(n)]
        else:
            smp.TotalParam[32] = [item / (3600 * 24 * 365.242) if index in rows else smp.TotalParam[32][index] for index, item in enumerate(stand_time_second)]  # stand year

    elif flag == 'smp':
        smp.TotalParam[67:71] = remove_none(smp.TotalParam[67:71], params[0:4], rows, n)
        smp.TotalParam[58:67] = remove_none(smp.TotalParam[58:67], params[4:13], rows, n)
        smp.TotalParam[97:100] = remove_none(smp.TotalParam[97:100], params[13:16], rows, n)
        smp.TotalParam[115:120] = remove_none(smp.TotalParam[115:120], params[16:21], rows, n)
        smp.TotalParam[126:136] = remove_none(smp.TotalParam[126:136], params[21:31], rows, n)
        # smp.TotalParam[120:123] = remove_none(smp.TotalParam[120:123], params[31:34], rows, n)
        smp.TotalParam[100:114] = remove_none(
            smp.TotalParam[100:114],
            [['Linear', 'Exponential', 'Power'][params[35:38].index(True)] if True in params[35:38] else '', *params[38:]], rows, n)
        pref = dict(zip(preference_keys, params[31:35]))
        smp.Info.preference.update(pref)
        for key, comp in get_components(smp).items():
            if isinstance(comp, Table):
                comp.decimal_places = pref['decimalPlaces']
                comp.set_coltypes()
        smp.AgeSpectraPlot.yaxis.title.text = f"Apparent Age ({str(pref['ageUnit']).capitalize()})"

    else:
        raise KeyError(f"{flag = } is not supported. It must be 'calc' for Calc Params, "
                       f"'irra' for Irradiation Params, or 'smp' for Sample Params.")
    return smp


def get_sequence(smp: Sample):
    set1_index = [index for index, _ in enumerate(smp.IsochronMark) if _ in [1, '1']]
    set2_index = [index for index, _ in enumerate(smp.IsochronMark) if _ in [2, '2']]
    set3_index = list(set(range(0, len(smp.IsochronMark))) - set(set1_index) - set(set2_index))
    smp.SelectedSequence1 = set1_index
    smp.SelectedSequence2 = set2_index
    smp.UnselectedSequence = set3_index
    smp.Info.results.selection[0]['data'] = smp.SelectedSequence1
    smp.Info.results.selection[1]['data'] = smp.SelectedSequence2
    smp.Info.results.selection[2]['data'] = smp.UnselectedSequence
    return ArArBasic(
        size=len(smp.SequenceName), name=smp.SequenceName, value=smp.SequenceValue, unit=smp.SequenceUnit,
        mark=ArArBasic(
            size=len(smp.IsochronMark), value=smp.IsochronMark,
            set1=ArArBasic(size=len(set1_index), index=set1_index),
            set2=ArArBasic(size=len(set2_index), index=set2_index),
            unselected=ArArBasic(size=len(set3_index), index=set3_index),
        )
    )


def get_results(smp: Sample):
    return ArArBasic(
        isochron=ArArBasic(**dict(
            ({'figure_2': 'normal', 'figure_3': 'inverse', 'figure_4': 'cl_1',
              'figure_5': 'cl_2', 'figure_6': 'cl_3', 'figure_7': 'three_d'}[key],
             ArArBasic(**dict(({2: 'unselected', 0: 'set1', 1: 'set2'}[_key],
                               ArArBasic(**_value)) for (_key, _value) in value.items())))
            for (key, value) in smp.Info.results.isochron.items())),
        age_plateau=ArArBasic(**dict(
            ({2: 'unselected', 0: 'set1', 1: 'set2'}[key], ArArBasic(**value))
            for (key, value) in smp.Info.results.age_plateau.items()))
    )




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - smp - table
# ==========================================
#
#
#
"""
import ast
import re
import copy
from .. import calc
from . import (sample as samples, basic)

Sample = samples.Sample
Table = samples.Table


# Table functions
def update_table_data(smp: Sample, only_table: str = None):
    """
    Update table data
    Parameters
    ----------
    smp
    only_table

    Returns
    -------

    """
    for key, comp in basic.get_components(smp).items():
        if not isinstance(comp, Table):
            continue
        if only_table is not None and key != only_table:
            continue
        if key == '1':
            data = calc.arr.merge(
                smp.SequenceName, smp.SequenceValue, *smp.SampleIntercept)
        elif key == '2':
            data = calc.arr.merge(
                smp.SequenceName, smp.SequenceValue, *smp.BlankIntercept)
        elif key == '3':
            data = calc.arr.merge(
                smp.SequenceName, smp.SequenceValue, *smp.CorrectedValues)
        elif key == '4':
            data = calc.arr.merge(
                smp.SequenceName, smp.SequenceValue, *smp.DegasValues)
        elif key == '5':
            smp.PublishValues[0] = copy.deepcopy(smp.DegasValues[ 0])
            smp.PublishValues[1] = copy.deepcopy(smp.DegasValues[ 8])
            smp.PublishValues[2] = copy.deepcopy(smp.DegasValues[10])
            smp.PublishValues[3] = copy.deepcopy(smp.DegasValues[20])
            smp.PublishValues[4] = copy.deepcopy(smp.DegasValues[24])
            smp.PublishValues[5:7] = copy.deepcopy(smp.ApparentAgeValues[2:4])
            smp.PublishValues[7:9] = copy.deepcopy(smp.ApparentAgeValues[6:8])
            data = calc.arr.merge(
                smp.SequenceName, smp.SequenceValue, *smp.PublishValues)
        elif key == '6':
            data = calc.arr.merge(
                smp.SequenceName, smp.SequenceValue, *smp.ApparentAgeValues)
        elif key == '7':
            data = calc.arr.merge(
                smp.SequenceName, smp.SequenceValue, smp.IsochronMark, *smp.IsochronValues)
        elif key == '8':
            data = calc.arr.merge(
                smp.SequenceName, smp.SequenceValue, *smp.TotalParam)
        else:
            data = [['']]
        # calc.arr.replace(data, pd.isnull, None)
        setattr(comp, 'data', calc.arr.transpose(data))


def update_handsontable(smp: Sample, data: list, id: str):
    """
    Parameters
    ----------
    smp : sample instance
    data : list
    id : str, table id

    Returns
    -------

    """

    def _normalize_data(a, cols, start_col=0):
        if len(a) >= cols:
            return a[start_col:cols]
        else:
            return a[start_col:] + [[''] * len(a[0])] * (cols - len(a))

    def _strToBool(cols):
        bools_dict = {
            'true': True, 'false': False, 'True': True, 'False': False, '1': True, '0': False, 'none': False,
        }
        return [bools_dict.get(str(col).lower(), False) for col in cols]

    def _digitize_data(a):
        # pattern = r'^[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?$'
        # return [[ast.literal_eval(str(cell)) if re.fullmatch(pattern, str(cell)) else np.nan if str(cell) == "" else cell for cell in row] for row in a]
        return a

    try:
        smp.SequenceName = data[0]
    except IndexError:
        pass

    update_all_table = False
    try:
        if data[1] != smp.SequenceValue:
            update_all_table = True
            smp.SequenceValue = data[1]
    except IndexError:
        pass

    if id == '1':  # 样品值
        data = _normalize_data(data, len(samples.SAMPLE_INTERCEPT_HEADERS), 2)
        smp.SampleIntercept = _digitize_data(data)
    elif id == '2':  # 本底值
        data = _normalize_data(data, len(samples.BLANK_INTERCEPT_HEADERS), 2)
        smp.BlankIntercept = _digitize_data(data)
    elif id == '3':  # 校正值
        data = _normalize_data(data, len(samples.CORRECTED_HEADERS), 2)
        smp.CorrectedValues = _digitize_data(data)
    elif id == '4':  # Degas table
        data = _normalize_data(data, len(samples.DEGAS_HEADERS), 2)
        smp.DegasValues = _digitize_data(data)
    elif id == '5':  # 发行表
        data = _normalize_data(data, len(samples.PUBLISH_TABLE_HEADERS), 2)
        smp.PublishValues = _digitize_data(data)
    elif id == '6':  # 年龄谱
        data = _normalize_data(data, len(samples.SPECTRUM_TABLE_HEADERS), 2)
        smp.ApparentAgeValues = _digitize_data(data)
    elif id == '7':  # 等时线
        smp.IsochronMark = _digitize_data(data)[2]
        data = _normalize_data(data, len(samples.ISOCHRON_TABLE_HEADERS), 3)
        smp.IsochronValues = _digitize_data(data)
        smp.sequence()
    elif id == '8':  # 总参数
        data = _normalize_data(data, len(samples.TOTAL_PARAMS_HEADERS), 2)
        data = _digitize_data(data)
        data[101: 112] = [_strToBool(i) for i in data[101: 112]]
        smp.TotalParam = data
    else:
        raise ValueError(f"{id = }, The table id is not supported.")
    if update_all_table:
        update_table_data(smp)
    else:
        update_table_data(smp, only_table=id)  # Update data of tables after changes of a table


def update_data_from_table(smp: Sample, only_table: str = None):
    """
    Update table data
    Parameters
    ----------
    smp
    only_table

    Returns
    -------

    """
    for key, comp in basic.get_components(smp).items():
        if not isinstance(comp, Table):
            continue
        if only_table is not None and key != only_table:
            continue
        if key == '1':
            smp.SampleIntercept = calc.arr.transpose(comp.data)[2:]
        elif key == '2':
            smp.BlankIntercept = calc.arr.transpose(comp.data)[2:]
        elif key == '3':
            smp.CorrectedValues = calc.arr.transpose(comp.data)[2:]
        elif key == '4':
            smp.DegasValues = calc.arr.transpose(comp.data)[2:]
        elif key == '5':
            smp.PublishValues = calc.arr.transpose(comp.data)[2:]
        elif key == '6':
            smp.ApparentAgeValues = calc.arr.transpose(comp.data)[2:]
        elif key == '7':
            smp.IsochronValues = calc.arr.transpose(comp.data)[3:]
        elif key == '8':
            smp.TotalParam = calc.arr.transpose(comp.data)[2:]
        else:
            pass



